import configparser
import os

from sha_learning.learning_setup.learner import ObsTable

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()


def save_data(symbols, distr, obstable: ObsTable, traces, time, sha_name, events_dict=None, path=None):
    SAVE_PATH = config['SUL CONFIGURATION']['REPORT_SAVE_PATH'].format(path)
    f = open(SAVE_PATH + sha_name + '.txt', 'w')
    content = ''
    # Report file structure:
    # [EVENT LABELS DICTIONARY]
    if events_dict is not None:
        content += '--EVENT LABELS DICT--\n\n'
        for pair in events_dict.items():
            content += '{}: {}\n'.format(pair[1], pair[0])

    # [MONITORED EVENTS AND SYMBOLS]
    content += '\n\n--OBSERVABLE EVENTS--\n\n'
    for sym in symbols.keys():
        content += '{}: {}'.format(sym, symbols[sym]) + '\n'

    # [LEARNED DISTRIBUTIONS]
    content += '\n\n--LEARNED DISTRIBUTIONS--\n\n'
    for i, d in enumerate(distr[0]):
        try:
            content += 'N_{}({:.6f}, {:.6f})\n'.format(i, d.params['avg'], d.params['var'])
        except KeyError:
            content += 'D_{}({:.6f})\n'.format(i, d.params['avg'])

    # [FINAL OBS TABLE]
    content += '\n\n--FINAL OBSERVATION TABLE--\n\n'
    content += obstable.__str__(filter_empty=True) + '\n'

    content += '\n\n--PERFORMANCE DATA--\n\n'
    # [SAMPLED TRACES]
    content += 'SAMPLED TRACES:\t{}\n'.format(traces)
    # [RUNNING TIME]
    content += 'RUNNING TIME:\t{}\n'.format(time)

    f.write(content)
    f.close()



from typing import List

from graphviz import Digraph

from sha_learning.domain.shafeatures import StochasticHybridAutomaton

FONT_OPEN_TAG = '<FONT {}>'
FONT_CLOSE_TAG = '</FONT>'
FONT_COLOR_ATTR = ' COLOR="{}"'
FONT_SIZE_ATTR = ' POINT-SIZE="{}"'


def style_label(args: List[str], font_sizes: List[int] = None, font_colors: List[str] = None):
    label = '<'
    for (index, arg) in enumerate(args):
        font_attrs = ''

        try:
            font_attrs += FONT_SIZE_ATTR.format(font_sizes[index])
        except (IndexError, TypeError):
            pass

        try:
            font_attrs += FONT_COLOR_ATTR.format(font_colors[index])
        except (IndexError, TypeError):
            pass

        label += FONT_OPEN_TAG.format(font_attrs)
        label += arg
        label += FONT_CLOSE_TAG
        label += '<br/>' if index != len(args) - 1 else ''

    label += '>'
    return label


def to_graphviz(ha: StochasticHybridAutomaton, name: str, SAVE_PATH: str, view=False):
    f = Digraph('hybrid_automaton', filename=SAVE_PATH + name)
    f.attr(rankdir='LR', size='2')
    f.attr('node', shape='circle')
    f.attr('node', fontname='helvetica')
    f.attr('node', penwidth='0.5')
    f.attr('edge', fontname='helvetica')
    f.attr('edge', penwidth='0.5')

    locations = ha.locations
    edges = ha.edges

    for loc in locations:
        if loc.flow_cond is not None:
            label = style_label([loc.name, '<br/><b>' + loc.flow_cond + '</b>'], [8, 6], ['black', '#ad0c00'])
        else:
            label = style_label([loc.name], [8, 6], ['black', '#ad0c00'])
        f.node(loc.name, label=label)

    for edge in edges:
        if edge.guard != '' and edge.sync != '':
            label = style_label([edge.guard, edge.sync], [8, 8], ['#008c05', '#0067b0'])
        elif edge.guard != '':
            label = style_label([edge.guard], [8], ['#008c05'])
        elif edge.sync != '':
            label = style_label([edge.sync], [8], ['#0067b0'])
        else:
            label = ''

        f.edge(edge.start.name, edge.dest.name, label=label)

    f.edge_attr.update(arrowsize='0.5')

    if view:
        f.view()

    return f



import configparser
import os
from typing import List, Set, Tuple

from pygad import GA

from sha_learning.case_studies.energy.sul_functions import label_event, parse_data, get_power_param, is_chg_pt
from sha_learning.domain.lshafeatures import Event, NormalDistribution, Trace
from sha_learning.domain.sigfeatures import Timestamp, SampledSignal
from sha_learning.domain.sulfeatures import SystemUnderLearning, RealValuedVar, FlowCondition

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

SPEED_RANGE = int(config['ENERGY CS']['SPEED_RANGE'])
MIN_SPEED = int(config['ENERGY CS']['MIN_SPEED'])
MAX_SPEED = int(config['ENERGY CS']['MAX_SPEED'])


def pwr_model(interval: List[Timestamp], P_0):
    interval = [ts.to_secs() for ts in interval]
    AVG_PW = 1.0
    return [AVG_PW] * len(interval)


# define flow conditions
on_fc: FlowCondition = FlowCondition(0, pwr_model)

# define distributions
off_distr = NormalDistribution(0, 0.0, 0.0)

model2distr = {0: []}
power = RealValuedVar([on_fc], [], model2distr, label='P')

# define events
events: List[Event] = []
for i in range(MIN_SPEED, MAX_SPEED, SPEED_RANGE):
    if i < MAX_SPEED - SPEED_RANGE:
        new_guard = '{}<=w<{}'.format(i, i + SPEED_RANGE)
    else:
        new_guard = '{}<=w'.format(i)
    events.append(Event(new_guard, 'start', 'm_{}'.format(len(events))))

spindle_off = Event('', 'stop', 'i_0')

events.append(spindle_off)
events.append(Event('', 'load', 'l'))
events.append(Event('', 'unload', 'u'))

DRIVER_SIG = ['w', 'pr']
DEFAULT_M = 0
DEFAULT_DISTR = 0

args = {'name': 'energy', 'driver': DRIVER_SIG, 'default_m': DEFAULT_M, 'default_d': DEFAULT_DISTR}
energy_cs = SystemUnderLearning([power], events, parse_data, label_event, get_power_param, is_chg_pt, args=args)

test = False
if test:
    TEST_PATH = config['TRACE GENERATION']['SIM_LOGS_PATH'].format('ENERGY')
    N = 10
    traces_files = os.listdir(TEST_PATH)
    traces_files = [file for file in traces_files if file[0] in ['_', 'W']]
    traces_files.sort()
    for file in traces_files:
        # testing data to signals conversion
        new_signals: List[SampledSignal] = parse_data(TEST_PATH + file)
        # testing chg pts identification
        chg_pts = energy_cs.find_chg_pts([sig for sig in new_signals if sig.label in DRIVER_SIG])
        # testing event labeling
        id_events = [label_event(events, new_signals, pt.t) for pt in chg_pts[:10]]
        # testing signal to trace conversion
        energy_cs.process_data(TEST_PATH + file)
        trace = energy_cs.timed_traces[-1]
        print('{}\t{}\t{}\t{}'.format(file, Trace(tt=trace),
                                      trace.t[-1].to_secs() - trace.t[0].to_secs(), len(trace)))
        # power_pts = new_signals[0].points
        # speed_pts = new_signals[1].points
        # pressure_pts = new_signals[2].points
        # sim_power = [0.0, 0.0, 8.2, 0.0, 3.12, 0.0, 2.4, 0.0, 1.5, 8.0, 0.0, 0.8, 0.0]
        # single_plot([pt.timestamp for pt in power_pts], [pt.value for pt in power_pts],
        #             energy_cs.timed_traces[-1].t, sim_power, trace)
        # double_plot([pt.timestamp for pt in power_pts], [pt.value for pt in power_pts],
        #           [pt.timestamp for pt in speed_pts], [pt.value for pt in speed_pts],
        #           trace, title=file, filtered=True,
        #           timestamps3=[pt.timestamp for pt in pressure_pts],
        #           v3=[pt.value for pt in pressure_pts])
        pass

    unique_events: List[Tuple[str, Set[Event]]] = []
    for i, t_trace in enumerate(energy_cs.timed_traces):
        trace = Trace(tt=t_trace)
        unique_events.append((traces_files[i], set(trace.events)))
    unique_events.sort(key=lambda s: len(s[1]), reverse=True)
    print(unique_events)

    unique_events.sort(key=lambda s: s[0])
    print(unique_events)


    def fitness_function(ga_instance: GA, solution, solution_idx):
        output = [unique_events[i] for i in solution]
        output_set = set()
        for tt in output:
            output_set = output_set.union(tt[1])
        return len(output_set) * sum([len(tt[1]) for tt in output])


    CLUSTER_DIM = 9

    ga_instance = GA(num_generations=50,
                     num_parents_mating=4,
                     fitness_func=fitness_function,
                     sol_per_pop=8,
                     num_genes=CLUSTER_DIM,
                     gene_type=int,
                     init_range_low=0,
                     init_range_high=len(energy_cs.timed_traces) - 1,
                     parent_selection_type="sss",
                     keep_parents=1,
                     crossover_type="single_point",
                     mutation_type="random",
                     mutation_percent_genes=10)

    ga_instance.run()

    solution, solution_fitness, solution_idx = ga_instance.best_solution()
    print("Parameters of the best solution : {solution}".format(solution=solution))
    print("Fitness value of the best solution = {solution_fitness}".format(solution_fitness=solution_fitness))

    # test segment identification
    # test_trace = Trace(energy_cs.traces[0][:1])
    # segments = energy_cs.get_segments(test_trace)
    #
    # # test model identification
    # TEACHER = Teacher(energy_cs)
    # identified_model: FlowCondition = TEACHER.mi_query(test_trace)
    # print(identified_model)
    #
    # # test distr identification
    # for i, trace in enumerate(TEACHER.timed_traces):
    #     for j, event in enumerate(trace.e):
    #         test_trace = Trace(energy_cs.traces[i][:j])
    #         identified_distr = TEACHER.ht_query(test_trace, identified_model, save=True)
    #
    #         segments = energy_cs.get_segments(test_trace)
    #         avg_metrics = sum([TEACHER.sul.get_ht_params(segment, identified_model)
    #                            for segment in segments]) / len(segments)
    #
    #         try:
    #             print('{}:\t{:.3f}->{}'.format(test_trace.events[-1].symbol, avg_metrics, identified_distr.params))
    #         except IndexError:
    #             print('{}:\t{:.3f}->{}'.format(test_trace, avg_metrics, identified_distr.params))
    #
    # for d in energy_cs.vars[0].distr:
    #     print(d.params)
    # distr_hist(TEACHER.hist)



import configparser
import csv
import os
from typing import List, Tuple, Dict

from sha_learning.domain.lshafeatures import Event, FlowCondition
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp, SignalPoint
from sha_learning.learning_setup.logger import Logger

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

try:
    CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
except ValueError:
    CS_VERSION = None
SPEED_RANGE = int(config['ENERGY CS']['SPEED_RANGE'])
PR_RANGE = int(config['ENERGY CS']['PR_RANGE'])
MIN_SPEED = int(config['ENERGY CS']['MIN_SPEED'])
MAX_SPEED = int(config['ENERGY CS']['MAX_SPEED'])

LOGGER = Logger('SUL DATA HANDLER')


def is_chg_pt(curr, prev):
    return abs(curr[0] - prev[0]) > SPEED_RANGE or abs(curr[1] - prev[1]) > PR_RANGE


def label_event(events: List[Event], signals: List[SampledSignal], t: Timestamp):
    speed_sig = signals[1]
    pressure_sig = signals[2]
    speed = {pt.timestamp: (i, pt.value) for i, pt in enumerate(speed_sig.points)}
    pressure = {pt.timestamp: (i, pt.value) for i, pt in enumerate(pressure_sig.points)}

    SPEED_INTERVALS: List[Tuple[int, int]] = []
    for i in range(MIN_SPEED, MAX_SPEED, SPEED_RANGE):
        if i < MAX_SPEED - SPEED_RANGE:
            SPEED_INTERVALS.append((i, i + SPEED_RANGE))
        else:
            SPEED_INTERVALS.append((i, None))

    curr_speed_index, curr_speed = speed[t]
    if curr_speed_index > 0:
        try:
            prev_index = [tup[0] for tup in speed.values() if tup[0] < curr_speed_index][-1]
            prev_speed = speed_sig.points[prev_index].value
        except IndexError:
            prev_speed = None
    else:
        prev_speed = curr_speed

    curr_press_index, curr_press = pressure[t]
    if curr_press_index > 0:
        try:
            prev_index = [tup[0] for tup in pressure.values() if tup[0] < curr_press_index][-1]
            prev_press = pressure_sig.points[prev_index].value
        except IndexError:
            prev_press = None
    else:
        prev_press = curr_press

    identified_event = None
    # if spindle was moving previously and now it is idle, return "stop" event
    if abs(curr_press - prev_press) > PR_RANGE:
        if curr_press > 500 and prev_press <= 500:
            identified_event = events[-2]
        else:
            identified_event = events[-1]
    elif curr_speed < MIN_SPEED and (prev_speed is not None and prev_speed >= MIN_SPEED):
        identified_event = events[-3]
    else:
        # if spindle is now moving at a different speed than before,
        # return 'new speed' event, which varies depending on current speed range
        if prev_speed is None or abs(curr_speed - prev_speed) >= SPEED_RANGE:
            for i, interval in enumerate(SPEED_INTERVALS):
                if (i < len(SPEED_INTERVALS) - 1 and interval[0] <= curr_speed < interval[1]) or \
                        (i == len(SPEED_INTERVALS) - 1 and curr_speed >= interval[0]):
                    identified_event = events[i]

    if identified_event is None:
        LOGGER.error("No event was identified at time {}.".format(t))

    return identified_event


def parse_ts(ts: str):
    date = ts.split(' ')[0].split('-')
    time = ts.split(' ')[1].split(':')
    return Timestamp(int(date[0]), int(date[1]), int(date[2]), int(time[0]), int(time[1]), int(time[2]))


def parse_data(path: str):
    # support method to parse traces sampled by ref query

    energy: SampledSignal = SampledSignal([], label='e')
    power: SampledSignal = SampledSignal([], label='P')
    speed: SampledSignal = SampledSignal([], label='w')
    pressure: SampledSignal = SampledSignal([], label='pr')

    with open(path) as csv_file:
        reader = csv.reader(csv_file, delimiter=',')

        prev_p = 0

        for i, row in enumerate(reader):
            if i > 0:
                ts = parse_ts(row[1])

                # parse energy value: if no reading is avaiable,
                # retrieve last available measurement
                try:
                    energy_v = float(row[2].replace(',', '.'))
                except ValueError:
                    energy_v = None
                energy.points.append(SignalPoint(ts, energy_v))

                # parse speed value: round to closest [100]
                try:
                    speed_v = round(float(row[3].replace(',', '.')) / 100) * 100
                    speed_v = max(speed_v, 0)
                except ValueError:
                    if len(speed.points) > 0:
                        speed_v = speed.points[-1].value
                    else:
                        speed_v = 0.0
                speed.points.append(SignalPoint(ts, speed_v))

                # parse pallet pressure value
                try:
                    pressure_v = float(row[4].replace(',', '.'))
                    prev_p = pressure_v
                except ValueError:
                    pressure_v = prev_p
                pressure.points.append(SignalPoint(ts, pressure_v))

        # transform energy signal into power signal by computing
        # the difference betwen two consecutive measurements
        power_pts: List[SignalPoint] = [SignalPoint(energy.points[0].timestamp, 0.0)]
        last_reading = energy.points[0].value
        last_power_value = 0.0
        for i, pt in enumerate(energy.points):
            if i == 0:
                continue
            elif pt.value is None:
                power_pts.append(SignalPoint(pt.timestamp, last_power_value))
            else:
                if last_reading is not None:
                    power_pts.append(SignalPoint(pt.timestamp, 60 * (pt.value - last_reading)))
                    last_power_value = 60 * (pt.value - last_reading)
                else:
                    power_pts.append(SignalPoint(pt.timestamp, 60 * (0.0)))
                    last_power_value = 0.0
                last_reading = pt.value
        power.points = power_pts

        # filter speed signal
        filtered_speed_pts: List[SignalPoint] = []
        last_switch = Timestamp(speed.points[0].timestamp.year, speed.points[0].timestamp.month,
                                speed.points[0].timestamp.day, speed.points[0].timestamp.hour,
                                speed.points[0].timestamp.min, 0)
        max_batch = speed.points[0].value
        timestamps: List[Timestamp] = []
        value_to_mul: Dict[float, int] = dict()
        for i, pt in enumerate(speed.points):
            curr_ts = Timestamp(pt.timestamp.year, pt.timestamp.month, pt.timestamp.day,
                                pt.timestamp.hour, pt.timestamp.min, 0)
            if curr_ts == last_switch and i < len(speed.points) - 1:
                timestamps.append(pt.timestamp)
                max_batch = min(max_batch, pt.value)
                if pt.value in value_to_mul:
                    value_to_mul[pt.value] += 1
                else:
                    value_to_mul[pt.value] = 1
            elif curr_ts != last_switch or i == len(speed.points) - 1:
                max_mul = 0
                max_val = 0
                for val in value_to_mul:
                    if value_to_mul[val] > max_mul:
                        max_mul = value_to_mul[val]
                        max_val = val
                for t in timestamps:
                    filtered_speed_pts.append(SignalPoint(t, max_val))
                last_switch = curr_ts
                max_batch = pt.value
                value_to_mul.clear()
                timestamps = [curr_ts]

        filtered_speed = SampledSignal(filtered_speed_pts, label='w')

        # infer pressure signal if absent
        non_zero_pts = [pt for pt in pressure.points if pt.value > 0]
        if len(non_zero_pts) == 0:
            inferred_pressure: List[SignalPoint] = []
            last_non_zero_speed_ts = filtered_speed.points[0].timestamp
            for pt in filtered_speed.points:
                delta = pt.timestamp.to_secs() - last_non_zero_speed_ts.to_secs()
                if pt.value > MIN_SPEED or delta < 600:
                    if pt.value > MIN_SPEED:
                        last_non_zero_speed_ts = pt.timestamp
                    pressure_v = 800
                else:
                    pressure_v = 0
                inferred_pressure.append(SignalPoint(pt.timestamp, pressure_v))
            pressure = SampledSignal(inferred_pressure, label='pr')
        pressure.points[0].value = 0
        pressure.points[-2].value = 0
        pressure.points[-1].value = 0
        for pt in filtered_speed.points[:20]:
            pt.value = 0.0
        filtered_speed.points[-2].value = 0.0

        return [power, filtered_speed, pressure]


def get_power_param(segment: List[SignalPoint], flow: FlowCondition):
    sum_power = sum([pt.value for pt in segment])
    avg_power = sum_power / len(segment)
    return avg_power



import configparser
import os
import warnings
from datetime import datetime

import sha_learning.pltr.lsha_report as report
import sha_learning.pltr.sha_pltr as sha_pltr
from sha_learning.case_studies.auto_twin.sul_definition import getSUL
from sha_learning.domain.lshafeatures import Trace
from sha_learning.domain.obstable import ObsTable
from sha_learning.domain.sulfeatures import SystemUnderLearning
from sha_learning.learning_setup.learner import Learner
from sha_learning.learning_setup.logger import Logger
from sha_learning.learning_setup.teacher import Teacher
from sha_learning.pltr.energy_pltr import distr_hist

warnings.filterwarnings('ignore')
startTime = datetime.now()

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS = 'AUTO_TWIN'
CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
RESAMPLE_STRATEGY = 'SKG'
LOGGER = Logger('LSHA')


def learn_automaton(pov: str, start_dt: str = None, end_dt: str = None,
                    start_ts: int = None, end_ts: int = None, save_path=None):
    auto_twin_cs, act_to_sensors = getSUL()
    SUL: SystemUnderLearning = auto_twin_cs
    SUL.reset_distributions()
    events_labels_dict = act_to_sensors

    TEACHER = Teacher(SUL, pov, start_dt, end_dt, start_ts, end_ts)

    long_traces = [Trace(events=[e]) for e in SUL.events]
    obs_table = ObsTable([], [Trace(events=[])], long_traces)
    LEARNER = Learner(TEACHER, obs_table)

    # RUN LEARNING ALGORITHM:
    LEARNED_SHA = LEARNER.run_lsha(filter_empty=True)

    # PLOT (AND SAVE) RESULT
    if save_path is None:
        SHA_SAVE_PATH = config['SUL CONFIGURATION']['SHA_SAVE_PATH'].format(
            os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning')
    else:
        SHA_SAVE_PATH = config['SUL CONFIGURATION']['SHA_SAVE_PATH'].format(save_path)

    SHA_NAME = '{}_{}_{}'.format(CS, RESAMPLE_STRATEGY, pov)
    if SHA_NAME in [file.split('-')[0] for file in os.listdir(SHA_SAVE_PATH)]:
        indexes = [file.split('-')[1][0] for file in os.listdir(SHA_SAVE_PATH) if SHA_NAME == file.split('-')[0]]
        max_index = max([int(x) for x in indexes])
        SHA_NAME += '-{}'.format(max_index + 1)
    else:
        SHA_NAME += '-0'
    graphviz_sha = sha_pltr.to_graphviz(LEARNED_SHA, SHA_NAME, SHA_SAVE_PATH, view=True)

    # saving sha source to .txt file
    sha_source = graphviz_sha.source
    with open(SHA_SAVE_PATH + SHA_NAME + '_source.txt', 'w') as f:
        f.write(sha_source)

    if config['DEFAULT']['PLOT_DISTR'] == 'True' and config['LSHA PARAMETERS']['HT_QUERY_TYPE'] == 'S':
        distr_hist(TEACHER.hist, SHA_NAME)

    report.save_data(TEACHER.symbols, TEACHER.distributions, LEARNER.obs_table,
                     len(TEACHER.signals), datetime.now() - startTime, SHA_NAME, events_labels_dict, save_path)
    LOGGER.info('----> EXPERIMENTAL RESULTS SAVED IN: {}{}.txt'.format(config['SUL CONFIGURATION']['REPORT_SAVE_PATH'],
                                                                       SHA_NAME))

    return SHA_NAME



import configparser
import os
import sys
import warnings
from datetime import datetime

import sha_learning.pltr.lsha_report as report
import sha_learning.pltr.sha_pltr as ha_pltr
from sha_learning.case_studies.auto_twin.sul_definition import auto_twin_cs, act_to_sensors
from sha_learning.case_studies.energy.sul_definition import energy_cs
from sha_learning.case_studies.energy_made.sul_definition import energy_made_cs
from sha_learning.case_studies.energy_sim.sul_definition import energy_sim_cs
from sha_learning.case_studies.hri.sul_definition import hri_cs
from sha_learning.case_studies.thermostat.sul_definition import thermostat_cs
from sha_learning.domain.lshafeatures import Trace
from sha_learning.domain.obstable import ObsTable
from sha_learning.domain.sulfeatures import SystemUnderLearning
from sha_learning.learning_setup.learner import Learner
from sha_learning.learning_setup.logger import Logger
from sha_learning.learning_setup.teacher import Teacher
from sha_learning.pltr.energy_pltr import distr_hist

# LEARNING PROCEDURE SETUP
warnings.filterwarnings('ignore')
startTime = datetime.now()

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS = config['SUL CONFIGURATION']['CASE_STUDY']
RESAMPLE_STRATEGY = config['SUL CONFIGURATION']['RESAMPLE_STRATEGY']
LOGGER = Logger('LSHA')

SUL: SystemUnderLearning
events_labels_dict = None
if CS == 'THERMO':
    SUL = thermostat_cs
elif CS == 'HRI':
    SUL = hri_cs
elif CS == 'ENERGY':
    if RESAMPLE_STRATEGY == 'SIM':
        SUL = energy_sim_cs
    elif RESAMPLE_STRATEGY == 'REAL':
        SUL = energy_cs
    elif RESAMPLE_STRATEGY == 'MADE':
        SUL = energy_made_cs
    else:
        raise RuntimeError
elif CS == 'AUTO_TWIN':
    SUL = auto_twin_cs
    events_labels_dict = act_to_sensors
else:
    raise RuntimeError

TEACHER = Teacher(SUL, pov=sys.argv[1], start_dt=sys.argv[2], end_dt=sys.argv[3])

long_traces = [Trace(events=[e]) for e in SUL.events]
obs_table = ObsTable([], [Trace(events=[])], long_traces)
LEARNER = Learner(TEACHER, obs_table)

# RUN LEARNING ALGORITHM:
LEARNED_HA = LEARNER.run_lsha(filter_empty=True)

# PLOT (AND SAVE) RESULT
HA_SAVE_PATH = config['SUL CONFIGURATION']['SHA_SAVE_PATH'].format(
    os.path.abspath(__file__).split('sha_learning')[0] + 'sha_learning/')

SHA_NAME = '{}_{}_{}'.format(CS, RESAMPLE_STRATEGY, config['SUL CONFIGURATION']['CS_VERSION'])
graphviz_sha = ha_pltr.to_graphviz(LEARNED_HA, SHA_NAME, HA_SAVE_PATH, view=True)

# saving sha source to .txt file
sha_source = graphviz_sha.source
with open(HA_SAVE_PATH.format(os.getcwd()) + SHA_NAME + '_source.txt', 'w') as f:
    f.write(sha_source)

if config['DEFAULT']['PLOT_DISTR'] == 'True' and config['LSHA PARAMETERS']['HT_QUERY_TYPE'] == 'S':
    distr_hist(TEACHER.hist, SHA_NAME)

report.save_data(TEACHER.symbols, TEACHER.distributions, LEARNER.obs_table,
                 len(TEACHER.signals), datetime.now() - startTime, SHA_NAME, events_labels_dict,
                 os.getcwd())
LOGGER.info('----> EXPERIMENTAL RESULTS SAVED IN: {}{}.txt'.format(config['SUL CONFIGURATION']['REPORT_SAVE_PATH'], SHA_NAME))



import configparser
import os

config = configparser.ConfigParser()
config.sections()
config.read('./resources/config/config.ini')
config.sections()

UPP_OUT_PATH = config['TRACE GENERATION']['UPPAAL_OUT_PATH']
CS = config['SUL CONFIGURATION']['CASE_STUDY']
CS_VERSION = config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', '')

folder = '/'.join(UPP_OUT_PATH.split('/')[:-1])
files = os.listdir(folder)
total = 0
for file in files:
    if file.startswith(CS.upper() + '_' + CS_VERSION):
        os.remove(folder + '/' + file)
        total += 1
print('{} traces deleted.'.format(total))



from enum import Enum
from typing import List


class LocLabels(Enum):
    IDLE = 'idle'
    BUSY = 'busy'


class Location:
    def __init__(self, name: str, flow_cond: str):
        self.name = name
        self.flow_cond = flow_cond

    def set_flow_cond(self, flow_cond: str):
        self.flow_cond = flow_cond

    def __eq__(self, other):
        same_name = self.name == other.name
        same_flow = self.flow_cond == other.flow_cond
        return same_name and same_flow

    def __str__(self):
        if self.flow_cond is not None:
            return self.name + ' ' + self.flow_cond
        else:
            return self.name

    def __hash__(self):
        return hash(str(self))


LOCATIONS: List[Location] = [Location(LocLabels.IDLE.value, None), Location(LocLabels.BUSY.value, None)]


class Edge:
    def __init__(self, start: Location, dest: Location, guard: str = '', sync: str = ''):
        self.start = start
        self.dest = dest
        self.guard = guard
        self.sync = sync

    def set_guard(self, guard):
        self.guard = guard

    def set_sync(self, sync):
        self.sync = sync

    def __eq__(self, other):
        same_start = self.start == other.start
        same_dest = self.dest == other.dest
        same_guard = self.guard == other.guard
        same_sync = self.sync == other.sync
        return same_start and same_dest and same_guard and same_sync


class StochasticHybridAutomaton:
    LOCATION_FORMATTER = 'q_{}'

    def __init__(self, loc: List[Location], edges: List[Edge]):
        self.locations = loc
        self.edges = edges

    def set_locations(self, loc: List[Location]):
        self.locations = loc

    def set_edges(self, edges: List[Edge]):
        self.edges = edges



import configparser
import math
import os
from typing import List

from sha_learning.case_studies.thermostat.sul_functions import label_event, parse_data, get_thermo_param, \
    is_chg_pt
from sha_learning.domain.lshafeatures import RealValuedVar, FlowCondition, Trace
from sha_learning.domain.sigfeatures import Event, Timestamp
from sha_learning.domain.sulfeatures import SystemUnderLearning
from sha_learning.learning_setup.teacher import Teacher

CLOSED_R = 100.0
OFF_DISTR = (100.0, 1.0, 200)
ON_DISTR = (0.7, 0.01, 200)
DRIVER_SIG = 't.ON'
DEFAULT_M = 1
DEFAULT_DISTR = 1

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()
CASE_STUDY = config['SUL CONFIGURATION']['CASE_STUDY']

try:
    CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
except ValueError:
    CS_VERSION = None


def off_model(interval: List[Timestamp], T_0: float):
    interval = [ts.to_secs() for ts in interval]
    return [T_0 * math.exp(-1 / OFF_DISTR[0] * (t - interval[0])) for t in interval]


def on_model(interval: List[Timestamp], T_0: float):
    interval = [ts.to_secs() for ts in interval]
    coeff = CLOSED_R * ON_DISTR[0]
    return [coeff - (coeff - T_0) * math.exp(-(1 / CLOSED_R) * (t - interval[0])) for t in interval]


def off_model_2(interval: List[Timestamp], T_0: float):
    interval = [ts.to_secs() for ts in interval]
    return [T_0 - 1 / OFF_DISTR[0] * (t - interval[0]) for t in interval]


def on_model_2(interval: List[Timestamp], T_0: float):
    interval = [ts.to_secs() for ts in interval]
    return [T_0 + ON_DISTR[0] * (t - interval[0]) for t in interval]


on_fc = FlowCondition(0, on_model)
off_fc = FlowCondition(1, off_model)

models: List[FlowCondition] = [on_fc, off_fc]

events = []

if CS_VERSION in [1]:
    on_event = Event('', 'on', 'h_0')
    off_event = Event('', 'off', 'c_0')
    events = [on_event, off_event]
if CS_VERSION in [2, 3, 4, 5, 6, 7, 8, 9, 10]:
    on_event = Event('!open', 'on', 'h_0')
    off_event = Event('!open', 'off', 'c_0')
    on_event2 = Event('open', 'on', 'h_1')
    off_event2 = Event('open', 'off', 'c_1')
    events = [on_event, off_event, on_event2, off_event2]
if CS_VERSION in [3, 8, 9, 10]:
    on_event3 = Event('open2', 'on', 'h_2')
    off_event3 = Event('open2', 'off', 'c_2')
    events += [on_event3, off_event3]
if CS_VERSION in [9, 10]:
    on_fc2 = FlowCondition(2, off_model_2)
    models += [on_fc2]
if CS_VERSION in [8]:
    on_fc2 = FlowCondition(2, on_model_2)
    off_fc2 = FlowCondition(3, off_model_2)
    models += [on_fc2, off_fc2]

model_to_distr = {}
for m in models:
    model_to_distr[m.f_id] = []

temperature = RealValuedVar(models, [], model_to_distr, label='T_r')
args = {'name': 'thermostat', 'driver': DRIVER_SIG, 'default_m': DEFAULT_M, 'default_d': DEFAULT_DISTR}
thermostat_cs = SystemUnderLearning([temperature], events, parse_data, label_event, get_thermo_param, is_chg_pt,
                                    args=args)

test = False
if test:
    # test event configuration
    print(thermostat_cs.symbols)

    # test trace processing
    thermo_traces = [file for file in os.listdir('./resources/traces/uppaal') if file.startswith('THERMO')]
    for trace in thermo_traces:
        thermostat_cs.process_data('./resources/traces/uppaal/' + trace)

    test_trace = Trace([on_event])
    plot_traces = [(i, t) for i, t in enumerate(thermostat_cs.traces) if t.startswith(test_trace)]

    # test visualization
    for tup in plot_traces[:1]:
        print(tup[1])
        thermostat_cs.plot_trace(index=tup[0], title='test', xlabel='time [s]', ylabel='degrees C°')

    # test segment identification
    segments = thermostat_cs.get_segments(test_trace)
    print(len(segments))

    # test model identification query
    teacher = Teacher(thermostat_cs)
    id_flow = teacher.mi_query(test_trace)
    print(id_flow)

    # test hypothesis testing query
    metrics = [get_thermo_param(s, id_flow) for s in segments]
    print(metrics)
    print(thermostat_cs.vars[0].model2distr[0])
    print(teacher.ht_query(Trace([on_event]), on_fc, save=True))
    print(thermostat_cs.vars[0].model2distr[0])
    print(thermostat_cs.vars[0].model2distr[1])
    print(teacher.ht_query(Trace([on_event, off_event]), off_fc, save=True))
    print(thermostat_cs.vars[0].model2distr[1])
    # thermostat_cs.plot_distributions()



import configparser
import os
import random
import subprocess
from typing import List, Set, Dict

import skg_main.skg_mgrs.connector_mgr as conn
from sha_learning.domain.lshafeatures import Trace, Event
from sha_learning.learning_setup.logger import Logger
from skg_main.skg_mgrs.skg_reader import Skg_Reader
from skg_main.skg_model.schema import Entity
from skg_main.skg_model.schema import Timestamp as skg_Timestamp
from skg_main.skg_model.semantics import EntityForest, EntityTree

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS = config['SUL CONFIGURATION']['CASE_STUDY']
CS_VERSION = config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', '')

UPP_EXE_PATH = config['TRACE GENERATION']['UPPAAL_PATH']
UPP_OUT_PATH = config['TRACE GENERATION']['UPPAAL_OUT_PATH']
SCRIPT_PATH = config['TRACE GENERATION']['UPPAAL_SCRIPT_PATH']

SIM_LOGS_PATH = config['TRACE GENERATION']['SIM_LOGS_PATH']
LOG_FILES = ['humanFatigue.log', 'humanPosition.log']

if CS == 'HRI':
    LINE_1 = ['bool force_exe = true;\n', 'bool force_exe']
    LINE_2 = ['int force_act[MAX_E] = ', 'int force_act']
    LINE_3 = ['const int TAU = {};\n', 'const int TAU']
    LINE_4 = ['amy = HFoll_{}(1, 48, 2, 3, -1);\n', 'amy = HFoll_']
    LINE_5 = ['const int VERSION = {};\n', 'const int VERSION']
    LINES_TO_CHANGE = [LINE_1, LINE_2, LINE_3, LINE_4, LINE_5]
else:
    LINE_1 = ['bool force_exe = true;\n', 'bool force_exe']
    LINE_2 = ['int force_open[MAX_E] = ', 'int force_open']
    LINE_3 = ['const int TAU = {};\n', 'const int TAU']
    LINE_4 = ['r = Room_{}(15.2);\n', 'r = Room']
    LINES_TO_CHANGE = [LINE_1, LINE_2, LINE_3, LINE_4]

UPP_MODEL_PATH = config['TRACE GENERATION']['UPPAAL_MODEL_PATH']
UPP_QUERY_PATH = config['TRACE GENERATION']['UPPAAL_QUERY_PATH'].format(CS_VERSION)

RESAMPLE_STRATEGY = config['SUL CONFIGURATION']['RESAMPLE_STRATEGY']

MAX_E = 15
LOGGER = Logger('TRACE GENERATOR')


class TraceGenerator:
    def __init__(self, word: Trace = Trace([]), pov: str = None,
                 start_dt: str = None, end_dt: str = None, start_ts: str = None, end_ts: str = None):
        self.word = word
        self.events: List[Event] = word.events
        self.evt_int: List[int] = []

        self.ONCE = False
        self.processed_traces: Set[str] = set()

        if RESAMPLE_STRATEGY == 'SKG':
            self.labels_hierarchy: List[List[str]] = []
            self.processed_entities: Dict[Entity, EntityTree] = {}
            self.pov = pov
            self.start_dt = start_dt
            self.end_dt = end_dt
            self.start_ts = start_ts
            self.end_ts = end_ts

    def set_word(self, w: Trace):
        self.events = w.events
        self.evt_int = []
        self.word = w

    def evts_to_ints(self):
        for e in self.events:
            if CS == 'HRI':
                if e.symbol in ['u_2', 'u_4']:
                    self.evt_int.append(1)
                elif e.symbol in ['u_3']:
                    self.evt_int.append(3)
                elif e.symbol in ['d_3', 'd_4']:
                    self.evt_int.append(0)
                elif e.symbol in ['d_2']:
                    self.evt_int.append(2)
                else:
                    self.evt_int.append(-1)
            else:
                # for thermo example: associates a specific value
                # to variable open for each event in the requested trace
                if int(CS_VERSION) < 8:
                    if e.symbol in ['h_0', 'c_0']:
                        self.evt_int.append(0)
                    elif e.symbol in ['h_1', 'c_1']:
                        self.evt_int.append(1)
                    elif e.symbol in ['h_2', 'c_2']:
                        self.evt_int.append(2)
                else:
                    if e.symbol in ['h_0', 'c_0']:
                        self.evt_int.append(0)
                    elif e.symbol in ['h_1', 'c_1']:
                        self.evt_int.append(1)
                    elif e.symbol in ['h_2', 'c_2']:
                        self.evt_int.append(2)
                    elif e.symbol in ['h_3', 'c_3']:
                        self.evt_int.append(0)

    def get_evt_str(self):
        self.evts_to_ints()

        res = '{'
        i = 0
        for evt in self.evt_int:
            res += str(evt) + ', '
            i += 1
        while i < MAX_E - 1:
            res += '-1, '
            i += 1
        res += '-1};\n'
        return res

    def fix_model(self):
        # customized uppaal model based on requested trace
        m_r = open(UPP_MODEL_PATH, 'r')

        new_line_1 = LINE_1[0]
        values = self.get_evt_str()
        new_line_2 = LINE_2[0] + values
        tau = max(len(self.evt_int) * 50, 200)
        new_line_3 = LINE_3[0].format(tau)
        new_line_4 = LINE_4[0].format(CS_VERSION)
        new_line_5 = LINE_5[0].format(int(CS_VERSION) - 1) if CS == 'HRI' else None
        new_lines = [new_line_1, new_line_2, new_line_3, new_line_4, new_line_5]

        lines = m_r.readlines()
        found = [False] * len(new_lines)
        for line in lines:
            for (i, l) in enumerate(LINES_TO_CHANGE):
                if line.startswith(LINES_TO_CHANGE[i][1]) and not found[i]:
                    lines[lines.index(line)] = new_lines[i]
                    found[i] = True
                    break

        m_r.close()
        m_w = open(UPP_MODEL_PATH, 'w')
        m_w.writelines(lines)
        m_w.close()

    def get_traces(self, n: int = 1):
        if RESAMPLE_STRATEGY == 'UPPAAL':
            return self.get_traces_uppaal(n)
        elif RESAMPLE_STRATEGY == 'SKG':
            return self.get_traces_skg(n)
        else:
            return self.get_traces_sim(n)

    def get_traces_skg(self, n: int = 1):
        driver = conn.get_driver()
        querier: Skg_Reader = Skg_Reader(driver)

        if len(self.labels_hierarchy) == 0:
            self.labels_hierarchy = querier.get_entity_labels_hierarchy()

        if self.start_ts is not None and self.end_ts is not None:
            START_T = int(self.start_ts)
            END_T = int(self.end_ts)
        else:
            def parse_date(s: str):
                fields = s.split('-')
                return skg_Timestamp(int(fields[0]), int(fields[1]), int(fields[2]), int(fields[3]), int(fields[4]),
                                     int(fields[5]))

            START_T = parse_date(self.start_dt)
            END_T = parse_date(self.end_dt)

        evt_seqs = []
        if self.pov.lower() == 'plant':
            pov = self.pov.lower()
            entity_tree = querier.get_entity_tree("Oven", EntityForest([]))
            events = querier.get_events_by_entity_tree_and_timestamp(entity_tree[0], START_T, END_T, pov)
            if len(events) > 0:
                evt_seqs.append(events)
        else:
            if self.pov.lower() == 'item':
                entities = querier.get_items(labels_hierarchy=self.labels_hierarchy, limit=n, random=True,
                                             start_t=START_T, end_t=END_T)
            else:
                entities = querier.get_resources(labels_hierarchy=querier.get_resource_labels_hierarchy(), limit=n, random=True)

            for entity in entities[:n]:
                if entity not in self.processed_entities:
                    pov = self.pov.lower()
                    if pov == 'item':
                        entity_tree = querier.get_entity_tree(entity.entity_id, EntityForest([]), reverse=True)
                        events = querier.get_events_by_entity_tree_and_timestamp(entity_tree[0], START_T, END_T,
                                                                                 pov)
                    else:
                        entity_tree = querier.get_entity_tree(entity.entity_id, EntityForest([]))
                        events = querier.get_events_by_entity_tree_and_timestamp(entity_tree[0], START_T, END_T,
                                                                                 pov)
                    if len(events) > 0:
                        evt_seqs.append(events)
                    self.processed_entities[entity] = entity_tree[0]

        conn.close_connection(driver)
        return evt_seqs

    def get_traces_sim(self, n: int = 1):
        # if self.ONCE:
        #    return []

        if CS.lower() == 'energy':
            sims = os.listdir(SIM_LOGS_PATH.format(CS))
            sims = list(filter(lambda s: s.startswith('_') and s not in self.processed_traces, sims))
            sims.sort()
        else:
            sims = os.listdir(SIM_LOGS_PATH.format(os.environ['RES_PATH'],
                                                   config['SUL CONFIGURATION']['CS_VERSION']))
            sims = list(filter(lambda s: s.startswith('SIM'), sims))
        paths = []
        for i in range(n + 1):
            if len(sims) == 0:
                break
            rand_sel = random.randint(0, 100)
            rand_sel = rand_sel % len(sims)
            if CS.lower() == 'energy':
                self.processed_traces.add(sims[rand_sel])
                paths.append(SIM_LOGS_PATH.format(CS) + '/' + sims[rand_sel])
            else:
                paths.append(SIM_LOGS_PATH.format(os.environ['RES_PATH'],
                                                  config['SUL CONFIGURATION']['CS_VERSION']) + '/' + sims[i] + '/')
        # self.ONCE = True
        return paths

    def get_traces_uppaal(self, n: int):
        # sample new traces through uppaal command line tool
        self.fix_model()
        LOGGER.debug('!! GENERATING NEW TRACES FOR: {} !!'.format(self.word))
        new_traces: List[str] = []

        for i in range(n):
            random.seed()
            n = random.randint(0, 2 ** 32)
            s = '{}_{}_{}'.format(CS, CS_VERSION, n)
            FNULL = open(os.devnull, 'w')
            p = subprocess.Popen([SCRIPT_PATH, UPP_EXE_PATH, UPP_MODEL_PATH,
                                  UPP_QUERY_PATH, str(n), UPP_OUT_PATH.format(s)], stdout=FNULL)
            p.wait()
            if p.returncode == 0:
                LOGGER.info('TRACES SAVED TO ' + s)
                # returns out file where new traces are stored
                new_traces.append(UPP_OUT_PATH.format(s))

        return new_traces



import configparser
import math
import os
from typing import List

from sha_learning.case_studies.hri.sul_functions import label_event, parse_data, get_ftg_param, \
    is_chg_pt
from sha_learning.domain.lshafeatures import RealValuedVar, FlowCondition, Trace
from sha_learning.domain.sigfeatures import Event, Timestamp
from sha_learning.domain.sulfeatures import SystemUnderLearning
from sha_learning.learning_setup.teacher import Teacher
from sha_learning.pltr.hri_pltr import double_plot

N_0 = (0.003, 0.0001, 100)
N_1 = (0.004, 0.0004, 100)

MAIN_SIGNAL = 0
DRIVER_SIG = '2'
DEFAULT_M = 0
DEFAULT_DISTR = 0

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()
CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', '')[0])
SAMPLE_STRATEGY = config['SUL CONFIGURATION']['RESAMPLE_STRATEGY']


def idle_model(interval: List[Timestamp], F_0: float):
    interval = [ts.to_secs() for ts in interval]
    return [F_0 * math.exp(-N_0[0] * (t - interval[0])) for t in interval]


def busy_model(interval: List[Timestamp], F_0: float):
    interval = [ts.to_secs() for ts in interval]
    return [1 - (1 - F_0) * math.exp(-N_1[0] * (t - interval[0])) for t in interval]


idle_fc = FlowCondition(0, idle_model)
busy_fc = FlowCondition(1, busy_model)

models: List[FlowCondition] = [idle_fc, busy_fc]

events = []
if SAMPLE_STRATEGY == 'UPPAAL':
    if CS_VERSION in [1]:
        u_event = Event('', 'start', 'u_1')
        d_event = Event('', 'stop', 'd_1')
        events = [u_event, d_event]
    elif CS_VERSION in [2]:
        u_event1 = Event('sit', 'start', 'u_1')
        d_event1 = Event('sit', 'stop', 'd_1')
        u_event2 = Event('!sit', 'start', 'u_2')
        d_event2 = Event('!sit', 'stop', 'd_2')
        events = [u_event1, d_event1, u_event2, d_event2]
    elif CS_VERSION in [3, 4, 5]:
        u_event1 = Event('sit!run', 'start', 'u_2')
        d_event1 = Event('sit!run', 'stop', 'd_2')
        u_event2 = Event('!sitrun', 'start', 'u_3')
        d_event2 = Event('!sitrun', 'stop', 'd_3')
        u_event3 = Event('!sit!run', 'start', 'u_4')
        d_event3 = Event('!sit!run', 'stop', 'd_4')
        events = [u_event1, d_event1, u_event2, d_event2, u_event3, d_event3]
else:
    if CS_VERSION in [1, 2, 3]:
        u_event1 = Event('sit!run', 'start', 'u_2')
        d_event1 = Event('sit!run', 'stop', 'd_2')
        u_event2 = Event('!sitrun', 'start', 'u_3')
        d_event2 = Event('!sitrun', 'stop', 'd_3')
        u_event3 = Event('!sit!run', 'start', 'u_4')
        d_event3 = Event('!sit!run', 'stop', 'd_4')
        events = [u_event1, d_event1, u_event2, d_event2, u_event3, d_event3]
    elif CS_VERSION in [4]:
        u_j = Event('!srh!l!a', 'start', 'u_j')
        u_n = Event('!sr!h!l!a', 'start', 'u_n')
        u_p = Event('!s!rhl!a', 'start', 'u_p')
        u_r = Event('!s!rh!l!a', 'start', 'u_r')
        u_t = Event('!s!r!hl!a', 'start', 'u_t')
        u_u = Event('!s!r!h!la', 'start', 'u_u')
        u_v = Event('!s!r!h!l!a', 'start', 'u_v')
        d_7 = Event('sr!h!l!a', 'stop', 'd_7')
        d_b = Event('s!rh!l!a', 'stop', 'd_b')
        d_f = Event('s!r!h!l!a', 'stop', 'd_f')
        d_j = Event('!srh!l!a', 'stop', 'd_j')
        d_p = Event('!s!rhl!a', 'stop', 'd_p')
        d_r = Event('!s!rh!l!a', 'stop', 'd_r')
        d_v = Event('!s!r!h!l!a', 'stop', 'd_v')
        events = [u_j, u_n, u_p, u_r, u_t, u_u, u_v, d_7, d_b, d_f, d_j, d_p, d_r, d_v]

model_to_distr = {}
for m in models:
    model_to_distr[m.f_id] = []

fatigue = RealValuedVar(models, [], model_to_distr, label='0')
args = {'name': 'hri', 'driver': DRIVER_SIG, 'default_m': DEFAULT_M, 'default_d': DEFAULT_DISTR}
hri_cs = SystemUnderLearning([fatigue], events, parse_data, label_event, get_ftg_param, is_chg_pt,
                             args=args)

test = False
if test:
    # test event configuration
    print(hri_cs.symbols)

    # test trace processing
    CS_VERSION = config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', '')
    TRACE_PATH = config['TRACE GENERATION']['SIM_LOGS_PATH'].format(os.environ['LSHA_RES_PATH'], CS_VERSION)
    hri_traces = [file for file in os.listdir(TRACE_PATH) if file.startswith('SIM')]
    hri_traces.sort()
    for trace in hri_traces:
        hri_cs.process_data(TRACE_PATH + '/' + trace + '/')
        print(hri_cs.traces[-1])

    test_trace = Trace([events[2], events[3]])
    plot_traces = [(i, t) for i, t in enumerate(hri_cs.traces) if t.startswith(test_trace)]

    # test segment identification
    segments = hri_cs.get_segments(test_trace)
    print(len(segments))

    # test model identification query
    teacher = Teacher(hri_cs)
    id_flow = teacher.mi_query(test_trace)
    print(id_flow)

    # test hypothesis testing query
    metrics = [get_ftg_param(s, id_flow) for s in segments]
    print(metrics)
    print(hri_cs.vars[0].model2distr[1])
    print(teacher.ht_query(test_trace, id_flow, save=True))

    for tup in plot_traces:
        double_plot(teacher.signals[tup[0]][0], teacher.signals[tup[0]][1], teacher.signals[tup[0]][3],
                    teacher.sul.timed_traces[tup[0]], str(tup[1]), events)



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy
import numpy as np


## reference: https://github.com/slaypni/fastdtw
## adapted from https://github.com/slaypni/fastdtw/blob/master/fastdtw/fastdtw.__dtw)
def est_derivatives(sig):
    '''
    Computing drivative differences between dx and dy
    Arguments:
           sig -- signal, numpy array of shape ( n,  )
    Result:
         d_sig -- estimated derivatives of input signal, numpy array of shape ( n-2,  )
    '''
    assert len(sig) >= 3, '''The length of your signal should be 
                           greater than 3 to implement DDTW.'''
    if type(sig) != numpy.ndarray:
        sig = np.array(sig)
    d_0 = sig[:-2]
    d_1 = sig[1:-1]
    d_2 = sig[2:]
    d_sig = ((d_1 - d_0) + (d_2 - d_0) / 2) / 2
    return d_sig


def generate_window(len_d_sig1, len_d_sig2, K):
    '''
    generate reduced search space
    Arguments:
           len_d_sig1 -- length of signal 1, scalar
           len_d_sig2 -- length of signal 2, scalar
                    K -- window size is 2 * K, scalar
    Result:
               window -- reduced search space, a python generator
    '''
    for i in range(len_d_sig1):
        lb = i - K
        ub = i + K
        if lb < 0 and ub < len_d_sig2:
            for j in range(ub):
                yield (i + 1, j + 1)
        elif lb >= 0 and ub < len_d_sig2:
            for j in range(lb, ub):
                yield (i + 1, j + 1)
        elif lb < 0 and ub >= len_d_sig2:
            for j in range(len_d_sig2):
                yield (i + 1, j + 1)
        elif lb >= 0 and ub >= len_d_sig2:
            for j in range(lb, len_d_sig2):
                yield (i + 1, j + 1)


def fast_ddtw(signal_1, signal_2, K=10):
    '''
    Arguments:
        signal_1 -- first time series, numpy array of shape ( n1,  )
        signal_2 -- second time series, numpy array of shape ( n2,  )
    Results:
        ddtw -- distance matrix, numpy array of shape ( n1 - 2, n2 - 2 )
        ddtw_traceback -- traceback matrix, numpy array of shape ( n1 - 2, n2 - 2 )
    '''

    d_sig1 = est_derivatives(signal_1)
    d_sig2 = est_derivatives(signal_2)

    len_d_sig1, len_d_sig2 = len(d_sig1), len(d_sig2)
    if K > abs(len_d_sig1 - len_d_sig2):
        window = generate_window(len_d_sig1, len_d_sig2, K)
    else:
        K = 2 * abs(len_d_sig1 - len_d_sig2)
        window = generate_window(len_d_sig1, len_d_sig2, K)
        print('input K is not a good choice... selected K = ', K, 'instead.')

    D = defaultdict(lambda: (float('inf'),))
    D[0, 0] = (0, 0, 0)
    for i, j in window:
        dt = abs(d_sig1[i - 1] - d_sig2[j - 1])
        D[i, j] = min((D[i - 1, j][0] + dt, i - 1, j), (D[i, j - 1][0] + dt, i, j - 1),
                      (D[i - 1, j - 1][0] + dt, i - 1, j - 1), key=lambda a: a[0])
    path = []
    i, j = len_d_sig1, len_d_sig2
    while not (i == j == 0):
        path.append((i - 1, j - 1))
        i, j = D[i, j][1], D[i, j][2]
    path.reverse()
    return (D[len_d_sig1, len_d_sig2][0], path)


def plot_raw_signals(signal_1, signal_2, title='raw_signals'):
    '''
    Arguments:
        signal_1 -- first time series, numpy array of shape ( n1,  )
        signal_2 -- second time series, numpy array of shape ( n2,  )
    Results:
          Figure
    '''
    plt.plot(signal_1)
    plt.plot(signal_2)
    plt.grid()
    plt.title(title)
    plt.xlabel('time')
    plt.ylabel('value')
    plt.show()


def plot_alignment_path(path, title='alignment_path'):
    '''
    Arguments:
          path -- aligned indices, list of indices
    Results:
        Figure
    '''
    plt.plot([index_pair[0] for index_pair in path], [index_pair[1] for index_pair in path])
    plt.grid()
    plt.title(title)
    plt.xlabel('signal 1')
    plt.ylabel('signal 2')
    plt.show()


def plot_aligned_signals(signal_1, signal_2, path, label1='real',
                         label2='theoretical', title='aligned_signals'):
    '''
    Arguments:
        signal_1 -- first time series, numpy array of shape ( n1,  )
        signal_2 -- second time series, numpy array of shape ( n2,  )
            path -- aligned indices, list of indices
    Results:
          Figure
    '''
    plt.plot([signal_1[index_pair[0]] for index_pair in path], label=label1)
    plt.plot([signal_2[index_pair[1]] for index_pair in path], label=label2)
    plt.legend()
    plt.grid()
    plt.title(title)
    plt.xlabel('time')
    plt.ylabel('value')
    plt.show()


def plot_actual_search_space(signal_1, signal_2, K):
    d_sig1 = est_derivatives(signal_1)
    d_sig2 = est_derivatives(signal_2)

    len_d_sig1, len_d_sig2 = len(d_sig1), len(d_sig2)
    if K > abs(len_d_sig1 - len_d_sig2):
        window = generate_window(len_d_sig1, len_d_sig2, K)
    else:
        K = 2 * abs(len_d_sig1 - len_d_sig2)
        window = generate_window(len_d_sig1, len_d_sig2, K)
        print('input K is not a good choice... selected K = ', K, 'instead.')
    ddtw_mat = np.zeros((len_d_sig1, len_d_sig2))
    for i, j in window:
        ddtw_mat[i - 1, j - 1] = 1
    ddtw_mat = ddtw_mat.T
    fig = plt.imshow(ddtw_mat[::-1, :])
    fig.axes.get_xaxis().set_visible(False)
    fig.axes.get_yaxis().set_visible(False)
    plt.show()


def plot_path_in_search_space(signal_1, signal_2, path, K):
    d_sig1 = est_derivatives(signal_1)
    d_sig2 = est_derivatives(signal_2)

    len_d_sig1, len_d_sig2 = len(d_sig1), len(d_sig2)
    if K > abs(len_d_sig1 - len_d_sig2):
        window = generate_window(len_d_sig1, len_d_sig2, K)
    else:
        K = 2 * abs(len_d_sig1 - len_d_sig2)
        window = generate_window(len_d_sig1, len_d_sig2, K)
        print('input K is not a good choice... selected K = ', K, 'instead.')
    ddtw_mat = np.zeros((len_d_sig1, len_d_sig2))
    for i, j in window:
        ddtw_mat[i - 1, j - 1] = 1
    for i, j in path:
        ddtw_mat[i, j] = 2
    ddtw_mat = ddtw_mat.T
    fig = plt.imshow(ddtw_mat[::-1, :])
    fig.axes.get_xaxis().set_visible(False)
    fig.axes.get_yaxis().set_visible(False)
    plt.show()



import configparser
import csv
import os
from typing import List, Tuple

from sha_learning.domain.lshafeatures import Event, FlowCondition
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp, SignalPoint
from sha_learning.learning_setup.logger import Logger

config = configparser.ConfigParser()  # open the configuration file
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

try:
    CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
except ValueError:
    CS_VERSION = None
SPEED_RANGE = int(config['ENERGY CS']['SPEED_RANGE'])
MIN_SPEED = int(config['ENERGY CS']['MIN_SPEED'])
MAX_SPEED = int(config['ENERGY CS']['MAX_SPEED'])

LOGGER = Logger('SUL DATA HANDLER')


def is_chg_pt(curr, prev):
    speed_cond = False
    if (curr[3] == 1 and prev[3] == 0) or (curr[3] == -1 and prev[3] == 0) or (curr[3] == 1 and prev[3] == -1):
        speed_cond = True
    return speed_cond or curr[1] != prev[1]


def label_event(events: List[Event], signals: List[SampledSignal], t: Timestamp):
    speed_sig = signals[1]
    pressure_sig = signals[2]
    speed = {pt.timestamp: (i, pt.value) for i, pt in enumerate(speed_sig.points)}
    pressure = {pt.timestamp: (i, pt.value) for i, pt in enumerate(pressure_sig.points)}

    # create a list of tuples where each tuple contains the limits of a speed interval
    SPEED_INTERVALS: List[Tuple[int, int]] = []
    for i in range(MIN_SPEED, MAX_SPEED, SPEED_RANGE):
        if i < MAX_SPEED - SPEED_RANGE:
            SPEED_INTERVALS.append((i, i + SPEED_RANGE))
        else:
            SPEED_INTERVALS.append((i, None))

    # identify the current and previous speed wrt the given timestamp t
    curr_speed_index, curr_speed = speed[t]
    if curr_speed_index > 0:
        try:
            prev_index = [tup[0] for tup in speed.values() if tup[0] < curr_speed_index][-1]
            prev_speed = speed_sig.points[prev_index].value
        except IndexError:
            prev_speed = None
    else:
        prev_speed = curr_speed

    # identify the current and previous pressure wrt the given timestamp t
    curr_press_index, curr_press = pressure[t]
    if curr_press_index > 0:
        try:
            prev_index = [tup[0] for tup in pressure.values() if tup[0] < curr_press_index][-1]
            prev_press = pressure_sig.points[prev_index].value
        except IndexError:
            prev_press = None
    else:
        prev_press = curr_press

    identified_event = None

    # if there is a pressure change, there is a load or unload event
    if curr_press != prev_press:
        # from 0 to 1 -> load
        if curr_press == 1.0 and prev_press == 0.0:
            identified_event = events[-2]
        # from 1 to 0 -> unload
        else:
            identified_event = events[-1]
    # if the previous velocity is bigger than the current one, we are going to 0 and we need to identify a stop event
    elif curr_speed < prev_speed:
        identified_event = events[-3]
    else:
        i = curr_speed_index
        while i < speed.__len__():
            const_speed = speed_sig.points[i].value
            if const_speed == speed_sig.points[i + 1].value and const_speed != 0:
                break
            else:
                i += 1
        # if the spindle is moving, return the constant speed that it will reach as a set point
        for i, interval in enumerate(SPEED_INTERVALS):
            if interval[0] <= const_speed <= interval[1]:
                identified_event = events[i]

    if identified_event is None:
        LOGGER.error("No event was identified at time {}.".format(t))

    return identified_event


def parse_ts(string: str):
    date = string.split('T')[0]
    year = int(date[0:4])
    month = int(date[5:7])
    day = int(date[8:10])
    time = string.split('T')[1].split('Z')[0]
    hour = int(time[0:2])
    minute = int(time[3:5])
    second = int(time[6:8])
    return Timestamp(year, month, day, hour, minute, second)


def parse_data(path: str):
    # support method to parse traces sampled by ref query
    power: SampledSignal = SampledSignal([], label='P')
    speed: SampledSignal = SampledSignal([], label='w')
    pressure: SampledSignal = SampledSignal([], label='pr')
    toolID: SampledSignal = SampledSignal([], label='id')
    speed_derivative: SampledSignal = SampledSignal([], label='wd')

    with open(path) as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')

        for i, row in enumerate(reader):
            # to skip the columns name row
            if i == 0:
                continue

            # select the timestamp from column 0
            ts = parse_ts(row['_time'])

            # parse power value
            power.points.append(SignalPoint(ts, float(row['Total_power'])))

            # parse speed value: round to closest [100]
            speed_v = round(float(row['actual_Speed_SP1']) / 100) * 100
            speed.points.append(SignalPoint(ts, speed_v))

            # parse pallet pressure value
            pressure.points.append(SignalPoint(ts, float(row['Pressure'])))

            # parse tool ID value
            toolID.points.append(SignalPoint(ts, float(row['dictID'])))

            # parse a signal which represents the derivative of the speed vector
            if i > 0:
                if round(speed.points[i - 2].value) == round(speed.points[i - 1].value):  # if constant
                    speed_d = 0
                elif round(speed.points[i - 2].value) < round(speed.points[i - 1].value):  # if going up
                    speed_d = 1
                else:  # if going down
                    speed_d = -1
                speed_derivative.points.append(SignalPoint(ts, float(speed_d)))

        return [power, speed, pressure, toolID, speed_derivative]


def get_power_param(segment: List[SignalPoint], flow: FlowCondition):
    sum_power = sum([pt.value for pt in segment])
    avg_power = sum_power / len(segment)
    return avg_power



from typing import List, Dict, Tuple

from sha_learning.domain.sigfeatures import Event, Timestamp

LOCATION_FORMATTER = 'q_{}'
EMPTY_STRING = '\u03B5'


class ProbDistribution:
    DISTR_FORMATTER = 'D_{}'

    def __init__(self, d_id: int, params: Dict[str, float]):
        self.params = params
        self.d_id = d_id
        self.label = self.DISTR_FORMATTER.format(d_id)
        self.d_id = d_id

    def __eq__(self, other):
        for p in self.params:
            if p not in other.params or self.params[p] != other.params[p]:
                return False
        return True

    def __str__(self):
        return self.label if self.params is not None else EMPTY_STRING


class NormalDistribution(ProbDistribution):
    DISTR_FORMATTER = 'N_{}'

    def __init__(self, d_id: int, avg: float, var: float):
        params = {'avg': avg, 'var': var}
        super().__init__(d_id, params)


class FlowCondition:
    MODEL_FORMATTER = 'f_{}'

    def __init__(self, f_id: int, f):
        self.f = f
        self.f_id = f_id
        self.label = self.MODEL_FORMATTER.format(f_id)

    def __eq__(self, other):
        return self.f_id == other.f_id

    def __str__(self):
        return self.label if self.f is not None else EMPTY_STRING

    def __hash__(self):
        return self.f_id


class RealValuedVar:
    def __init__(self, flows: List[FlowCondition], distr: List[ProbDistribution],
                 m2d: Dict[int, List[int]], label: str = None):
        self.flows = flows
        self.distr = distr
        self.model2distr = m2d
        self.label = label

    def __eq__(self, other):
        same_flows = all([f in other.flows for f in self.flows])
        same_distr = all([d in other.distr for d in self.distr])
        return same_flows and same_distr

    def get_distr_for_flow(self, x: int):
        related_distr = self.model2distr[x]
        return list(filter(lambda d: d.d_id in related_distr, self.distr))


class TimedTrace:
    def __init__(self, t: List[Timestamp], e: List[Event]):
        self.t = t
        self.e = e

    def __eq__(self, other):
        return all([ts == other.t[i] and self.e[i] == other.e[i] for i, ts in enumerate(self.t)])

    def __len__(self):
        return len(self.t)


class Trace:
    def __init__(self, events: List[Event] = None, tt: TimedTrace = None):
        if tt is not None:
            self.events = tt.e
        else:
            self.events = events

    def __str__(self):
        if len(self.events) == 0:
            return EMPTY_STRING
        else:
            return ''.join([e.symbol for e in self.events])

    def __eq__(self, other):
        return str(self) == str(other)

    def __len__(self):
        return len(self.events)

    def __getitem__(self, item):
        return self.events[item]

    def __add__(self, other):
        return Trace(events=self.events + other.events)

    def sub_prefix(self, prefix):
        return Trace(events=self.events[len(prefix.events):])

    def __hash__(self):
        return hash(str(self))

    def get_prefixes(self):
        prefixes: List[Trace] = []
        for i in range(len(self)):
            if i == 0:
                prefixes.append(Trace([self[0]]))
            else:
                prefixes.append(Trace(self[:i + 1]))
        return prefixes

    def startswith(self, word):
        if len(word) > len(self):
            return False

        for i, e in enumerate(word):
            if self.events[i].symbol != e.symbol:
                return False
        else:
            return True


class State:
    def __init__(self, vars: List[Tuple[FlowCondition, ProbDistribution]]):
        self.vars = vars
        self.label = ''
        for i, v in enumerate(vars):
            if v[0] is not None and v[1] is not None:
                self.label += '({},{})'.format(v[0], v[1])
            else:
                self.label += '({},{})'.format(EMPTY_STRING, EMPTY_STRING)
            if i < len(vars) - 1:
                self.label += ';'

    def __str__(self):
        return self.label

    def __eq__(self, other):
        return self.label == other.label

    def __hash__(self):
        return hash(self.label)

    def observed(self):
        return any([pair[0] is not None and pair[1] is not None for pair in self.vars])



import configparser
import os
from typing import List, Dict, Tuple

from sha_learning.domain.lshafeatures import Event, FlowCondition
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp, SignalPoint
from sha_learning.learning_setup.logger import Logger

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

try:
    CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
except ValueError:
    CS_VERSION = None

LOGGER = Logger('SUL DATA HANDLER')

POV = config['AUTO-TWIN CONFIGURATION']['POV'].lower()


def is_chg_pt(curr, prev):
    return curr[0] != prev[0] and curr[0] > 0.0


def label_event(events: List[Event], signals: List[SampledSignal], t: Timestamp):
    curr_value = [pt.value for pt in signals[0].points if pt.timestamp == t][0]

    identified_event = [e for e in events if int(e.symbol.replace('s', '')) == int(curr_value)][0]

    return identified_event


def parse_ts(ts):
    try:
        return Timestamp(ts.year, ts.month, ts.day, ts.hour, ts.mins, int(ts.sec))
    except AttributeError:
        try:
            return Timestamp(ts.year, ts.month, ts.day, ts.hour, ts.minute, ts.second)
        except AttributeError:
            return Timestamp(0, 0, 0, 0, 0, ts)


# FIXME should be generic
ACT_TO_SENSORS = {"Entrada Material Sucio": 'S1', "Cargado en carro  L+D": 'S2',
                  "Carga L+D iniciada": 'S3', "Carga L+D liberada": 'S4',
                  "Montaje": 'S5', "Producción  montada": 'S6', "Comisionado": 'S10',
                  "Composición de cargas": 'S7', "Carga de esterilizador liberada": 'S8',
                  "Carga de esterilizadorliberada": 'S9', 'Pass Sensor S1': 'S1', 'Pass Sensor S2': 'S2',
                  'Pass Sensor S3': 'S3', 'Pass Sensor S4': 'S4', 'Pass Sensor S106': 'S106',
                  'Pass Sensor S5': 'S5', 'Pass Sensor S6': 'S6', 'Pass Sensor S101': 'S101',
                  'Pass Sensor S105': 'S105', 'Pass Sensor S100': 'S100', 'Pass Sensor S7': 'S7',
                  'Pass Sensor S8': 'S8', 'Pass Sensor S102': 'S102', 'Pass Sensor S104': 'S104',
                  'Pass Sensor S9': 'S9', 'Pass Sensor S10': 'S10', 'Pass Sensor S103': 'S103',
                  'Pass Sensor S11': 'S11', 'Pass Sensor S12': 'S12', 'Pass Sensor S13': 'S13',
                  'Pass Sensor S14': 'S14', 'Pass Sensor S15': 'S15',
                  'Pass Sensor S16': 'S16', 'Pass Sensor S17': 'S17',
                  'Start Break': 'S200', 'Stop Break': 'S201',
                  'Read Lock Status': 'S202', 'Read WIP amount': 'S203', 'Pass Sensor LOAD_1': 'S11',
                  'Pass Sensor PROCESS_1': 'S12', 'Pass Sensor UNLOAD_1': 'S13',
                  'Pass Sensor LOAD_2': 'S21', 'Pass Sensor PROCESS_2': 'S22', 'Pass Sensor UNLOAD_2': 'S23',
                  'Pass Sensor FAIL_1': 'S14', 'Pass Sensor BLOCK_2': 'S24', 'Pass Sensor LOAD_3': 'S31',
                  'Pass Sensor PROCESS_3': 'S32', 'Pass Sensor LOAD_4': 'S41', 'Pass Sensor UNLOAD_3': 'S33',
                  'Pass Sensor PROCESS_4': 'S42', 'Pass Sensor LOAD_5': 'S51', 'Pass Sensor PROCESS_5': 'S52',
                  'Pass Sensor UNLOAD_5': 'S53', 'Pass Sensor UNLOAD_4': 'S43', 'Pass Sensor FAIL_5': 'S54',
                  'Pass Sensor BLOCK_1': 'S15', 'Pass Sensor BLOCK_5': 'S55', 'Pass Sensor BLOCK_3': 'S34',
                  'Pass Sensor BLOCK_4': 'S44', 'Pass Sensor CS001': 'S1', 'Pass Sensor CS002': 'S2',
                  'Pass Sensor CS003': 'S3',
                  'Pass Sensor CS004': 'S4', 'Pass Sensor CS005': 'S5', 'Pass Sensor CS006': 'S6',
                  'Pass Sensor CS101': 'S101', 'Pass Sensor CS105': 'S105', 'Pass Sensor CS100': 'S100',
                  'Pass Sensor CS007': 'S7', 'Pass Sensor CS008': 'S8', 'Pass Sensor CS102': 'S102',
                  'Pass Sensor CS104': 'S104', 'Pass Sensor CS009': 'S9', 'Pass Sensor CS010': 'S10',
                  'Pass Sensor CS4102': 'S4102', 'Pass Sensor CS4103': 'S4103',
                  'Pass Sensor CS4104': 'S4104', 'Pass Sensor CS4201': 'S4201',
                  'Pass Sensor CS4202': 'S4202', 'Pass Sensor CS4301': 'S4301',
                  'Pass Sensor CS4401': 'S4401', 'Pass Sensor CS106': 'S106', 'Pass Sensor CS011': 'S11',
                  'Pass Sensor CS012': 'S12', 'Pass Sensor CS013': 'S13', 'Pass Sensor CS014': 'S14',
                  'Pass Sensor CS015': 'S15', 'Pass Sensor CS016': 'S16', 'Pass Sensor CS4101': 'S4101',
                  'Pass Sensor CS4203': 'S4203', 'Pass Sensor CS4204': 'S4204', 'Pass Sensor VS001': 'S501',
                  'Pass Sensor VS002': 'S502', 'Pass Sensor VS003': 'S503',
                  'Read power': 'S601', 'Read temperature': 'S602',
                  "Dirty Material Input": "S1",
                  "Assembled production": "S2",
                  "Composition of charges": "S3",
                  "Sterilizer Load Released": "S4",
                  "Commissioner": "S5",
                  "Load L+D released": "S6",
                  "Assembly": "S7",
                  "Loaded on L+D trolley": "S8",
                  "L+D Load Started": "S9",
                  "Enter Cleaning": "S10",
                  "Enter Storage": "S11",
                  "Exit Storage": "S12"}


def update_state_vector(path, state_vector: List[int], sensor_to_station: Dict[str, Tuple[int, str]]):
    if len(path) == 0:
        return
    else:
        last_sensor = ACT_TO_SENSORS[path[0].activity]
        if sensor_to_station[last_sensor][1] is None:
            return update_state_vector(path[1:], state_vector, sensor_to_station)
        elif sensor_to_station[last_sensor][1] == 'IN':
            state_vector[sensor_to_station[last_sensor][0]] += 1
        else:
            state_vector[sensor_to_station[last_sensor][0]] -= 1
        return update_state_vector(path[1:], state_vector, sensor_to_station)


def vec_to_base_x(vector: List[int], x: int):
    res = 0
    for i in range(0, len(vector)):
        res += vector[len(vector) - 1 - i] * x ** i
    return res


def parse_value(path, i):
    if path[i].activity not in ACT_TO_SENSORS:
        s_id = float(int(path[i].activity.replace('S', '')))
    else:
        sensor = ACT_TO_SENSORS[path[i].activity]
        s_id = float(int(sensor.replace('S', '')))

    if POV == 'plant':
        # determine resource state vector
        # TODO this should become system-agnostic
        if os.environ['NEO4J_SCHEMA'] == 'pizzaLineV1':
            state_vector = [0, 0, 0, 0, 0]
            sensor_to_station = {'S1': (0, None), 'S2': (1, 'IN'), 'S3': (1, 'OUT'),
                                 'S7': (2, None), 'S4': (3, 'IN'), 'S5': (3, 'OUT'), 'S6': (4, None)}
        elif os.environ['NEO4J_SCHEMA'] == 'pizzaLineV2':
            state_vector = [0] * 10
            sensor_to_station = {'S1': (0, None), 'S2': (1, 'IN'), 'S3': (1, 'OUT'),
                                 'S16': (2, None), 'S4': (3, 'IN'), 'S5': (3, 'OUT'),
                                 'S6': (4, 'IN'), 'S7': (4, 'OUT'), 'S8': (5, 'IN'), 'S9': (5, 'OUT'),
                                 'S10': (6, 'IN'), 'S11': (6, 'OUT'), 'S12': (6, 'OUT'),
                                 'S17': (7, None), 'S13': (8, 'OUT'), 'S14': (8, 'OUT'),
                                 'S15': (9, None)}
        else:
            state_vector = [0] * 13
            sensor_to_station = {'S1': (0, None), 'S202': (0, None),
                                 'S2': (1, 'IN'), 'S3': (1, 'OUT'), 'S100': (2, None),
                                 'S4': (3, 'IN'), 'S5': (3, 'OUT'),
                                 'S6': (4, 'IN'), 'S7': (4, 'OUT'), 'S105': (4, 'OUT'), 'S101': (5, None),
                                 'S8': (6, 'IN'), 'S9': (6, 'OUT'), 'S104': (6, 'OUT'), 'S102': (7, None),
                                 'S10': (8, 'IN'), 'S11': (8, 'OUT'), 'S12': (8, 'OUT'),
                                 'S103': (9, None), 'S16': (10, None),
                                 'S13': (11, 'IN'), 'S14': (11, 'OUT'), 'S203': (11, None), 'S15': (12, None)}

        update_state_vector(path[:i + 1], state_vector, sensor_to_station)
        idle_busy_vector = [int(v > 0) for v in state_vector]

        return s_id, vec_to_base_x(idle_busy_vector, 2)
    else:
        return s_id


def parse_data(path):
    sensor_id: SampledSignal = SampledSignal([], label='s_id')
    sensor_id.points.append(SignalPoint(Timestamp(0, 0, 0, 0, 0, 0), 0))
    DELTA_T = 100000
    if POV == 'plant':
        state_signal: SampledSignal = SampledSignal([], label='state_vec')
        state_signal.points.append(SignalPoint(Timestamp(0, 0, 0, 0, 0, 0), 0))
    for i, ekg_event in enumerate(path):
        if ekg_event.date is None:
            ts = parse_ts(ekg_event.timestamp)
            if i < len(path) - 1:
                next_ts = parse_ts(path[i + 1].timestamp)
                new_tss = [Timestamp.from_secs(t) for t in range(ts.to_secs(), next_ts.to_secs(), DELTA_T)]
            else:
                new_tss = [ts]
        else:
            ts = parse_ts(ekg_event.date)
            if i < len(path) - 1:
                next_ts = parse_ts(path[i + 1].date)
                new_tss = [Timestamp.from_secs(t) for t in range(ts.to_secs(), next_ts.to_secs(), DELTA_T)]
            else:
                new_tss = [ts]

        # FIXME: this should be generic.
        if POV == 'plant':
            value, value_v = parse_value(path, i)
        else:
            value = parse_value(path, i)

        if i > 0 and ts == parse_ts(path[i - 1].timestamp):
            # in case there are two events at the same time, the last one overrides.
            sensor_id.points[-1].value = value
            if POV == 'plant':
                state_signal.points[-1].value = value_v
        elif len(new_tss) > 1:
            sensor_id.points.extend([SignalPoint(t, value) for t in new_tss[:-1]])
            sensor_id.points.append(SignalPoint(new_tss[-1], 0.0))
            if POV == 'plant':
                state_signal.points.extend([SignalPoint(t, value_v) for t in new_tss])
        elif len(new_tss) > 0:
            sensor_id.points.append(SignalPoint(new_tss[-1], value))
            if POV == 'plant':
                state_signal.points.append(SignalPoint(new_tss[-1], value_v))
            if i < len(path) - 1:
                next_ts = parse_ts(path[i + 1].timestamp)
                sensor_id.points.append(SignalPoint(Timestamp.from_secs(next_ts.to_secs() - 1), 0.0))
                if POV == 'plant':
                    state_signal.points.append(SignalPoint(Timestamp.from_secs(next_ts.to_secs() - 1), value_v))
        else:
            sensor_id.points.append(SignalPoint(ts, value))
            if POV == 'plant':
                state_signal.points.append(SignalPoint(ts, value_v))

    last_ts = sensor_id.points[-1].timestamp
    sensor_id.points.append(
        SignalPoint(Timestamp(last_ts.year, last_ts.month, last_ts.day, last_ts.hour, last_ts.min, last_ts.sec + 1),
                    sensor_id.points[-1].value))

    if POV == 'plant':
        state_signal.points.append(
            SignalPoint(Timestamp(last_ts.year, last_ts.month, last_ts.day, last_ts.hour, last_ts.min, last_ts.sec + 1),
                        state_signal.points[-1].value))

        return [sensor_id, state_signal]
    else:
        return [sensor_id]


def get_rand_param(segment: List[SignalPoint], flow: FlowCondition):
    return segment[0].value



import configparser
import math
import os
from typing import List

from sha_learning.domain.lshafeatures import FlowCondition
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp, Event, SignalPoint
from sha_learning.learning_setup.logger import Logger

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

try:
    CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
except ValueError:
    CS_VERSION = None

ON_R = 100.0
LOGGER = Logger('SUL DATA HANDLER')


def is_chg_pt(curr, prev):
    return curr != prev


def label_event(events: List[Event], signals: List[SampledSignal], t: Timestamp):
    heatOn = signals[0]
    wOpen = signals[2]
    t = t.to_secs()

    identified_guard = ''
    curr_wOpen = list(filter(lambda x: x.timestamp.to_secs() <= t, wOpen.points))[-1]
    if CS_VERSION in [2, 4, 5, 6, 7]:
        identified_guard += events[0].guard if curr_wOpen.value == 1.0 else events[2].guard
    if CS_VERSION in [3, 8, 9, 10]:
        if curr_wOpen.value == 2.0:
            identified_guard += events[4].guard
        elif curr_wOpen.value == 1.0:
            identified_guard += events[2].guard
        else:
            identified_guard += events[0].guard

    curr_heatOn = list(filter(lambda x: x.timestamp.to_secs() <= t, heatOn.points))[-1]
    identified_channel = events[0].chan if curr_heatOn.value == 1.0 else events[1].chan

    identified_event = [e for e in events if e.guard == identified_guard and e.chan == identified_channel][0]
    return identified_event


def parse_data(path: str):
    # support method to parse traces sampled by ref query
    f = open(path, 'r')
    variables = ['t.ON', 'T_r', 'r.open']
    lines = f.readlines()
    split_indexes = [lines.index(k + ':\n') for k in variables]
    split_lines = [lines[i + 1:split_indexes[ind + 1]] for ind, i in enumerate(split_indexes) if
                   i != split_indexes[-1]]
    split_lines.append(lines[split_indexes[-1] + 1:len(lines)])
    new_signals: List[SampledSignal] = []
    for i, v in enumerate(variables):
        entries = [line.split(' ') for line in split_lines[i]][0][1:]
        entries = [entry.replace('(', '').replace(')', '').replace('\n', '') for entry in entries]
        t = [float(x.split(',')[0]) for x in entries]
        values = [float(x.split(',')[1]) for x in entries]
        ts = [Timestamp(0, 0, 0, 0, 0, i) for i in t]
        signal_pts: List[SignalPoint] = [SignalPoint(ts[i], values[i]) for i in range(len(ts))]
        new_signal: SampledSignal = SampledSignal(signal_pts, v)
        new_signals.append(new_signal)
    return new_signals


def get_thermo_param(segment: List[SignalPoint], flow: FlowCondition):
    try:
        val = [pt.value for pt in segment]
        if flow.f_id in [0, 2]:
            if CS_VERSION in [1, 2, 3, 4, 5, 6, 7] or (CS_VERSION in [8, 9, 10] and flow.f_id in [0]):
                increments = []
                for (i, pt) in enumerate(val):
                    if i > 0 and pt != val[i - 1]:
                        increments.append(pt - val[i - 1] * math.exp(-1 / ON_R))
                Ks = [delta_t / (ON_R * (1 - math.exp(-1 / ON_R))) for delta_t in increments if delta_t != 0]

                LOGGER.info('Estimating rate with heat on ({})'.format(flow.f_id))
                est_rate = sum(Ks) / len(Ks) if len(Ks) > 0 else None
            else:
                increments = [pt - val[i - 1] for i, pt in enumerate(val) if i > 0]
                increments = [i for i in increments if i != 0]
                LOGGER.info('Estimating rate with heat on ({})'.format(flow.f_id))
                est_rate = sum(increments) / len(increments) if len(increments) > 0 else None
        else:
            if CS_VERSION in [1, 2, 3, 4, 5, 6, 7] or (CS_VERSION in [8, 9, 10] and flow.f_id in [1]):
                increments = []
                for (i, pt) in enumerate(val):
                    if i > 0:
                        increments.append(pt / val[i - 1])

                Rs = [-1 / math.log(delta_t) for delta_t in increments if delta_t != 1]

                LOGGER.info('Estimating rate with heat off ({})'.format(flow.f_id))
                est_rate = sum(Rs) / len(Rs) if len(Rs) > 0 else None
            else:
                increments = [pt - val[i - 1] for i, pt in enumerate(val) if i > 0]
                increments = [x for x in increments if x != 0]
                increments = [-1 / x for x in increments]
                LOGGER.info('Estimating rate with heat off ({})'.format(flow.f_id))
                est_rate = sum(increments) / len(increments) if len(increments) > 0 else None
        return est_rate
    except ValueError:
        return None



from typing import List, Dict

import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats

from sha_learning.domain.lshafeatures import Trace, TimedTrace, RealValuedVar, FlowCondition, ProbDistribution
from sha_learning.domain.sigfeatures import ChangePoint, Event, SampledSignal, Timestamp, SignalPoint


class SystemUnderLearning:
    def __init__(self, rv_vars: List[RealValuedVar], events: List[Event],
                 parse_f, label_f, param_f, is_chg_pt, **args):
        #
        self.vars = rv_vars
        self.flows = [v.flows for v in rv_vars]
        self.events = events
        self.symbols = {e.symbol: e.label for e in events}
        self.parse_f = parse_f
        self.label_f = label_f
        self.param_f = param_f
        self.is_chg_pt = is_chg_pt
        #
        self.signals: List[List[SampledSignal]] = []
        self.timed_traces: List[TimedTrace] = []
        self.traces: List[Trace] = []

        self.name = args['args']['name']
        self.driver = args['args']['driver']
        self.default_m = args['args']['default_m']
        self.default_d = args['args']['default_d']

    def add_distribution(self, d: ProbDistribution, f: FlowCondition):
        self.vars[0].distr.append(d)
        self.vars[0].model2distr[f.f_id].append(d.d_id)

    def reset_distributions(self):
        for i, v in enumerate(self.vars):
            self.vars[i].distr = []

    #
    # TRACE PROCESSING METHODS
    #
    def find_chg_pts(self, driver: List[SampledSignal]):
        values = [{pt.timestamp: pt.value for pt in sig.points} for sig in driver]
        chg_pts: List[ChangePoint] = []

        # IDENTIFY CHANGE PTS IN DRIVER OVERLAY
        prev = [sig.points[0].value for sig in driver]
        for ts in [pt.timestamp for pt in driver[0].points]:
            curr = [val_dic[ts] for val_dic in values]
            if self.is_chg_pt(curr, prev):
                chg_pts.append(ChangePoint(ts))
            prev = curr

        return chg_pts

    def process_data(self, path):
        new_signals: List[SampledSignal] = self.parse_f(path)
        self.signals.append(new_signals)

        driver_sig = [sig for sig in new_signals if sig.label in self.driver]

        chg_pts = self.find_chg_pts(driver_sig)
        events = [self.label_f(self.events, new_signals, pt.t) for pt in chg_pts]
        new_tt = TimedTrace([pt.t for pt in chg_pts], events)
        self.timed_traces.append(new_tt)
        self.traces.append(Trace(tt=new_tt))

    def get_ht_params(self, segment: List[SignalPoint], flow: FlowCondition):
        return self.param_f(segment, flow)

    def get_segments(self, word: Trace):
        traces: List[int] = [i for i, t in enumerate(self.traces) if t.startswith(word)]
        if len(traces) == 0:
            return []

        segments = []
        # for all traces, get signal segment from last(word) to the following event
        for trace in traces:
            main_sig_index = [i for i, s in enumerate(self.signals[0]) if s.label == self.vars[0].label][0]
            main_sig = self.signals[trace][main_sig_index]

            if word != '' and str(word) != 'ε':
                start_timestamp = self.timed_traces[trace].t[max(len(word) - 1, 0)].to_secs()
            else:
                start_timestamp = Timestamp(0, 0, 0, 0, 0, 0).to_secs()

            if len(word) < len(self.timed_traces[trace]):
                end_timestamp = self.timed_traces[trace].t[len(word)].to_secs()
            else:
                end_timestamp = main_sig.points[-1].timestamp.to_secs()

            segment = [pt for pt in main_sig.points if start_timestamp <= pt.timestamp.to_secs() < end_timestamp]
            segments.append(segment)
        else:
            return segments

    #
    # VISUALIZATION METHODS
    #
    def plot_trace(self, index, title=None, xlabel=None, ylabel=None):
        plt.figure(figsize=(50, 5))

        if title is not None:
            plt.title(title, fontsize=18)
        if xlabel is not None:
            plt.xlabel(xlabel, fontsize=18)
        if ylabel is not None:
            plt.ylabel(ylabel, fontsize=18)

        signal_to_plot = [i for i, s in enumerate(self.signals[0]) if s.label == self.vars[0].label][0]
        tt = self.timed_traces[index]

        sig = self.signals[index][signal_to_plot].points
        t = [pt.timestamp.to_secs() for pt in sig]
        v = [pt.value for pt in sig]

        plt.xlim(min(t[1:]) - 5, max(t) + 5)
        plt.ylim(0, max(v) + .05)
        plt.plot(t, v, 'k', linewidth=.5)

        plt.vlines([ts.to_secs() for ts in tt.t], [0] * len(tt), [max(v)] * len(tt), 'b', '--')
        for index, evt in enumerate(tt.e):
            plt.text(tt.t[index].to_secs() - 7, max(v) + .01, str(evt), fontsize=18, color='blue')

        plt.show()

    def plot_distributions(self):
        for flow in self.flows[0]:
            plt.figure()
            plt.title("Distributions for {}".format(flow.label))
            related_distributions = self.vars[0].get_distr_for_flow(flow.f_id)
            for d in related_distributions:
                distr: Dict[str, float] = d.params
                mu: float = distr['avg']
                sigma: float = distr['var']
                x = np.linspace(mu - 3 * sigma, mu + 3 * sigma, 200)
                plt.plot(x, stats.norm.pdf(x, mu, sigma), label='{}({:.6f}, {:.6f})'.format(d, mu, sigma))
            plt.legend()
            plt.show()



import configparser
import os
from typing import List, Tuple, Dict, Set

from sha_learning.domain.lshafeatures import State, FlowCondition, ProbDistribution
from sha_learning.domain.obstable import ObsTable, Row, Trace
from sha_learning.domain.shafeatures import StochasticHybridAutomaton, Location, Edge
from sha_learning.learning_setup.logger import Logger
from sha_learning.learning_setup.teacher import Teacher

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

LOGGER = Logger('LEARNER')


class Learner:
    def __init__(self, teacher: Teacher, table: ObsTable):
        self.symbols = teacher.symbols
        self.TEACHER = teacher
        default_table = table
        self.obs_table = table if table is not None else default_table

    def fill_row(self, row: Row, i: int, s_word: Trace, obs: List[Row]):
        for j, t_word in enumerate(self.obs_table.get_E()):
            # if cell is yet to be filled,
            # asks teacher to answer queries
            # and fills cell with answers
            if not obs[i].state[j].observed():
                identified_model: FlowCondition = self.TEACHER.mi_query(s_word + t_word)
                identified_distr: ProbDistribution = self.TEACHER.ht_query(s_word + t_word, identified_model)
                if identified_model is None or identified_distr is None:
                    identified_model = None
                    identified_distr = None
                row.state[j] = State([(identified_model, identified_distr)])
                if not State([(identified_model, identified_distr)]).observed() and j == 0:
                    break
        return row

    def fill_table(self):
        upp_obs: List[Row] = self.obs_table.get_upper_observations()
        for i, s_word in enumerate(self.obs_table.get_S()):
            row: Row = Row(upp_obs[i].state.copy())
            row = self.fill_row(row, i, s_word, upp_obs)
            upp_obs[i] = row
        self.obs_table.set_upper_observations(upp_obs)

        low_obs: List[Row] = self.obs_table.get_lower_observations()
        for i, s_word in enumerate(self.obs_table.get_low_S()):
            row: Row = Row(low_obs[i].state.copy())
            row = self.fill_row(row, i, s_word, low_obs)
            low_obs[i] = row
        self.obs_table.set_lower_observations(low_obs)

    def is_closed(self):
        upp_obs: List[Row] = self.obs_table.get_upper_observations()
        low_obs: List[Row] = self.obs_table.get_lower_observations()
        for l_i, row in enumerate(low_obs):
            if not row.is_populated():
                continue
            row_is_in_upper = False
            for s_i, s_word in enumerate(self.obs_table.get_S()):
                if self.TEACHER.eqr_query(row, upp_obs[s_i]):
                    row_is_in_upper = True
                    break
            if not row_is_in_upper:
                return False
        else:
            return True

    def is_consistent(self):
        events = self.TEACHER.sul.events
        upp_obs = self.obs_table.get_upper_observations()
        pairs: List[Tuple[Trace, Trace]] = []
        # FIXME: each pair shows up twice, duplicates should be cleared
        for index, row in enumerate(upp_obs):
            equal_rows = [i for (i, r) in enumerate(upp_obs) if
                          index != i and self.TEACHER.eqr_query(row, r, strict=True)]
            S = self.obs_table.get_S()
            equal_pairs = [(S[index], S[equal_i]) for equal_i in equal_rows]
            pairs += equal_pairs
        if len(pairs) == 0:
            return True, None
        else:
            for pair in pairs:
                for e in events:
                    try:
                        new_pair_1 = self.obs_table.get_S().index(pair[0] + Trace([e]))
                        new_row_1 = self.obs_table.get_upper_observations()[new_pair_1]
                    except ValueError:
                        new_pair_1 = self.obs_table.get_low_S().index(pair[0] + Trace([e]))
                        new_row_1 = self.obs_table.get_lower_observations()[new_pair_1]

                    try:
                        new_pair_2 = self.obs_table.get_S().index(pair[1] + Trace([e]))
                        new_row_2 = self.obs_table.get_upper_observations()[new_pair_2]
                    except ValueError:
                        new_pair_2 = self.obs_table.get_low_S().index(pair[1] + Trace([e]))
                        new_row_2 = self.obs_table.get_lower_observations()[new_pair_2]

                    rows_different = not self.TEACHER.eqr_query(new_row_1, new_row_2)
                    if new_row_1.is_populated() and new_row_2.is_populated() and rows_different:
                        for e_i, e_word in enumerate(self.obs_table.get_E()):
                            if new_row_1.state[e_i] != new_row_2.state[e_i] \
                                    and (Trace([e]) + e_word) not in self.obs_table.get_E():
                                LOGGER.warn('INCONSISTENCY: {}-{}'.format(pair[0] + Trace([e]), pair[1] + Trace([e])))
                                return False, Trace([e]) + e_word
            else:
                return True, None

    def make_closed(self):
        upp_obs: List[Row] = self.obs_table.get_upper_observations()
        low_S: List[Trace] = self.obs_table.get_low_S()
        low_obs: List[Row] = self.obs_table.get_lower_observations()
        for index, row in enumerate(low_obs):
            # if there is a populated row in lower portion that is not in the upper portion
            # the corresponding word is added to the S word set
            row_present = any([self.TEACHER.eqr_query(row, row_2) for row_2 in upp_obs])
            if row.is_populated() and not row_present:
                upp_obs.append(Row(row.state))
                new_s_word: Trace = low_S[index]
                self.obs_table.add_S(new_s_word)
                low_obs.pop(index)
                self.obs_table.del_low_S(index)
                # lower portion is then updated with all combinations of
                # new S word and all possible symbols
                for event in self.TEACHER.sul.events:
                    self.obs_table.add_low_S(new_s_word + Trace([event]))
                    empty_state = State([(None, None)])
                    new_row: Row = Row([empty_state] * len(self.obs_table.get_E()))
                    low_obs.append(new_row)
        self.obs_table.set_upper_observations(upp_obs)
        self.obs_table.set_lower_observations(low_obs)

    def make_consistent(self, discr_sym: Trace):
        self.obs_table.add_E(discr_sym)
        upp_obs: List[Row] = self.obs_table.get_upper_observations()
        low_obs: List[Row] = self.obs_table.get_lower_observations()
        for s_i in range(len(upp_obs)):
            upp_obs[s_i].state.append(State([(None, None)]))
        for s_i in range(len(low_obs)):
            low_obs[s_i].state.append(State([(None, None)]))

    def add_counterexample(self, counterexample: Trace):
        upp_obs = self.obs_table.get_upper_observations()
        low_obs = self.obs_table.get_lower_observations()

        # add counterexample and all its prefixes to S
        for i, e in enumerate(counterexample):
            new_trace = Trace(counterexample[:i + 1])
            if new_trace not in self.obs_table.get_S():
                if new_trace in self.obs_table.get_low_S():
                    to_copy: Row = low_obs[self.obs_table.get_low_S().index(new_trace)]
                else:
                    to_copy: Row = Row([State([(None, None)])] * len(self.obs_table.get_E()))
                self.obs_table.get_S().append(new_trace)
                upp_obs.append(Row(to_copy.state))

            if new_trace in self.obs_table.get_low_S():
                row_index = self.obs_table.get_low_S().index(new_trace)
                self.obs_table.get_lower_observations().pop(row_index)
                self.obs_table.get_low_S().pop(row_index)

            # add 1-step away words to low_S
            for a in self.TEACHER.sul.events:
                if new_trace + Trace([a]) not in self.obs_table.get_low_S() \
                        and new_trace + Trace([a]) not in self.obs_table.get_S():
                    self.obs_table.get_low_S().append(new_trace + Trace([a]))
                    new_state: List[State] = []
                    # add empty cells to T
                    for j in range(len(self.obs_table.get_E())):
                        new_state.append(State([(None, None)]))
                    low_obs.append(Row(new_state))

        self.obs_table.set_upper_observations(upp_obs)
        self.obs_table.set_lower_observations(low_obs)

    @staticmethod
    def get_nondetermistic_edge(sha: StochasticHybridAutomaton, loc: Location):
        outgoing_edges = [e for e in sha.edges if e.start == loc]
        seen_events: Set[str] = set()
        for edge in outgoing_edges:
            if edge.sync in seen_events:
                LOGGER.warn('NON-DETERMINISM DETECTED! Location: {}, Event: {}'.format(loc.name, edge.sync))
                return edge.sync
            else:
                seen_events.add(edge.sync)
        return None

    def merge_loc(self, sha: StochasticHybridAutomaton, loc: Location,
                  event: str, loc_dic: Dict[Trace, str]):
        competing_locs = [edge.dest for edge in sha.edges if edge.start == loc and edge.sync == event]
        # If ambiguous rows have different flow/distr, they cannot be merged.
        if len(set([l.flow_cond for l in competing_locs])) > 1:
            return sha, False

        ambig_traces = [tr for tr in loc_dic.keys() for l in competing_locs if loc_dic[tr] == l.name]
        ambig_rows = [row for i, row in enumerate(self.obs_table.get_upper_observations()) for tr in ambig_traces if
                      self.obs_table.get_S().index(tr) == i]

        # if two among the competing rows are not at least weakly equal,
        # merging is not possible
        checked_pairs: List[Tuple[Row, Row]] = []
        for i, row_1 in enumerate(ambig_rows):
            for j, row_2 in enumerate(ambig_rows):
                if i != j and (row_1, row_2) not in checked_pairs and (row_2, row_1) not in checked_pairs:
                    if not self.TEACHER.eqr_query(row_1, row_2):
                        return sha, False
                    else:
                        checked_pairs.append((row_1, row_2))

        # otherwise, merge.
        to_remove = [e for e in sha.edges if e.dest in competing_locs[1:] and e.start == loc]
        starting_in_competing = [e for e in sha.edges if e.start in competing_locs[1:]]
        ending_in_competing = [e for e in sha.edges if e.dest in competing_locs[1:] and e.start != loc]
        for e in to_remove:
            sha.edges.remove(e)
        for e in starting_in_competing:
            sha.edges.append(Edge(competing_locs[0], e.dest, e.guard, e.sync))
            sha.edges.remove(e)
        for e in ending_in_competing:
            sha.edges.append(Edge(e.start, competing_locs[0], e.guard, e.sync))
            try:
                sha.edges.remove(e)
            except ValueError:
                LOGGER.warn('Edge already removed.')

        for l in competing_locs[1:]:
            try:
                sha.locations.remove(l)
            except ValueError:
                LOGGER.warn('Location already removed.')

        return sha, True

    def sanity_check(self, sha: StochasticHybridAutomaton, loc_dic: Dict[Trace, str]):
        to_check: Set[Location] = set(sha.locations)

        while len(to_check) > 0:
            for loc in sha.locations:
                non_det_event = Learner.get_nondetermistic_edge(sha, loc)
                if non_det_event is not None:
                    sha, merged = self.merge_loc(sha, loc, non_det_event, loc_dic)
                    if merged:
                        to_check = set(sha.locations)
                        break
                    else:
                        LOGGER.warn('MERGING LOCATIONS UNSUCCESSFUL.')
                        to_check = set()
                        break
                else:
                    to_check.remove(loc)

        return sha

    def run_lsha(self, debug_print=True, filter_empty=False):
        # Fill Observation Table with Answers to Queries (from TEACHER)
        step0 = True  # work around to implement a do-while structure

        self.fill_table()

        self.TEACHER.ref_query(self.obs_table)
        self.fill_table()

        counterexample = self.TEACHER.get_counterexample(self.obs_table)

        while counterexample is not None or step0:
            step0 = False
            # if a counterexample was found, it (and all of its prefixes) are added to set S
            if counterexample is not None:
                LOGGER.warn('FOUND COUNTEREXAMPLE: {}'.format(counterexample))
                self.add_counterexample(counterexample)
                self.fill_table()

            if debug_print:
                LOGGER.info('OBSERVATION TABLE')
                self.obs_table.print(filter_empty)

            # Check if obs. table is closed
            closedness_check = self.is_closed()
            consistency_check, discriminating_symbol = self.is_consistent()
            while not (closedness_check and consistency_check):
                if not closedness_check:
                    LOGGER.warn('!!TABLE IS NOT CLOSED!!')
                    # If not, make closed
                    self.make_closed()
                    self.fill_table()
                    LOGGER.info('CLOSED OBSERVATION TABLE')
                    self.obs_table.print(filter_empty)

                # Check if obs. table is consistent
                if not consistency_check:
                    LOGGER.warn('!!TABLE IS NOT CONSISTENT!!')
                    # If not, make consistent
                    self.make_consistent(discriminating_symbol)
                    self.fill_table()
                    LOGGER.info('CONSISTENT OBSERVATION TABLE')
                    self.obs_table.print(filter_empty)

                closedness_check = self.is_closed()
                consistency_check, discriminating_symbol = self.is_consistent()

            self.TEACHER.ref_query(self.obs_table)
            self.fill_table()
            if debug_print:
                LOGGER.info('OBSERVATION TABLE')
                self.obs_table.print(filter_empty)
            counterexample = self.TEACHER.get_counterexample(self.obs_table)

        if debug_print:
            LOGGER.info('FINAL OBSERVATION TABLE')
            self.obs_table.print(filter_empty)
        # Build Hypothesis Automaton
        LOGGER.info('BUILDING HYP. AUTOMATON...')
        hypsha, loc_dict = self.obs_table.to_sha(self.TEACHER)
        try:
            hypsha = self.sanity_check(hypsha, loc_dict)
        except:
            LOGGER.error("Error occurred while fixing the SHA.")
        return hypsha



import configparser
import csv
import os
from typing import List, Tuple

from sha_learning.domain.lshafeatures import Event, FlowCondition
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp, SignalPoint
from sha_learning.learning_setup.logger import Logger

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

try:
    CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', ''))
except ValueError:
    CS_VERSION = None
SPEED_RANGE = int(config['ENERGY CS']['SPEED_RANGE'])
MIN_SPEED = int(config['ENERGY CS']['MIN_SPEED'])
MAX_SPEED = int(config['ENERGY CS']['MAX_SPEED'])

LOGGER = Logger('SUL DATA HANDLER')


def is_chg_pt(curr, prev):
    return abs(curr[0] - prev[0]) > SPEED_RANGE or curr[1] != prev[1]


def label_event(events: List[Event], signals: List[SampledSignal], t: Timestamp):
    speed_sig = signals[1]
    pressure_sig = signals[2]
    speed = {pt.timestamp: (i, pt.value) for i, pt in enumerate(speed_sig.points)}
    pressure = {pt.timestamp: (i, pt.value) for i, pt in enumerate(pressure_sig.points)}

    SPEED_INTERVALS: List[Tuple[int, int]] = []
    for i in range(MIN_SPEED, MAX_SPEED, SPEED_RANGE):
        if i < MAX_SPEED - SPEED_RANGE:
            SPEED_INTERVALS.append((i, i + SPEED_RANGE))
        else:
            SPEED_INTERVALS.append((i, None))

    curr_speed_index, curr_speed = speed[t]
    if curr_speed_index > 0:
        try:
            prev_index = [tup[0] for tup in speed.values() if tup[0] < curr_speed_index][-1]
            prev_speed = speed_sig.points[prev_index].value
        except IndexError:
            prev_speed = None
    else:
        prev_speed = curr_speed

    curr_press_index, curr_press = pressure[t]
    if curr_press_index > 0:
        try:
            prev_index = [tup[0] for tup in pressure.values() if tup[0] < curr_press_index][-1]
            prev_press = pressure_sig.points[prev_index].value
        except IndexError:
            prev_press = None
    else:
        prev_press = curr_press

    identified_event = None

    if curr_press != prev_press:
        if curr_press == 1.0 and prev_press == 0.0:
            identified_event = events[-2]
        else:
            identified_event = events[-1]
    # if spindle was moving previously and now it is idle, return "stop" event
    elif curr_speed < MIN_SPEED and (prev_speed is not None and prev_speed >= MIN_SPEED):
        identified_event = events[-3]
    else:
        # if spindle is now moving at a different speed than before,
        # return 'new speed' event, which varies depending on current speed range
        if prev_speed is None or abs(curr_speed - prev_speed) >= SPEED_RANGE:
            for i, interval in enumerate(SPEED_INTERVALS):
                if (i < len(SPEED_INTERVALS) - 1 and interval[0] <= curr_speed < interval[1]) or \
                        (i == len(SPEED_INTERVALS) - 1 and curr_speed >= interval[0]):
                    identified_event = events[i]

    if identified_event is None:
        LOGGER.error("No event was identified at time {}.".format(t))

    return identified_event


def parse_ts(ts: str):
    fields = ts.split(':')
    return Timestamp(0, 0, 0, int(fields[0]), int(fields[1]), int(fields[2]))


def parse_data(path: str):
    # support method to parse traces sampled by ref query
    power: SampledSignal = SampledSignal([], label='P')
    speed: SampledSignal = SampledSignal([], label='w')
    pressure: SampledSignal = SampledSignal([], label='pr')

    with open(path) as csv_file:
        reader = csv.reader(csv_file, delimiter=',')
        counter = 0

        for i, row in enumerate(reader):
            if i == 0:
                continue

            ts = parse_ts(row[2])

            if i > 1 and ts == speed.points[-1].timestamp:
                # parse power value
                power.points[-1].value = (power.points[-1].value * counter + float(row[4])) / (counter + 1)

                # parse speed value: round to closest [100]
                speed_v = round(float(row[3]) / 100) * 100
                speed.points[-1].value = min(speed_v, speed.points[-1].value)

                # parse pallet pressure value
                pressure_v = float(row[1] != 'UNLOAD')
                pressure.points[-1].value = min(pressure_v, pressure.points[-1].value)

                counter += 1
            else:
                counter = 0

                # parse power value
                power.points.append(SignalPoint(ts, float(row[4])))

                # parse speed value: round to closest [100]
                speed_v = round(float(row[3]) / 100) * 100
                speed.points.append(SignalPoint(ts, speed_v))

                # parse pallet pressure value
                pressure_v = float(not (row[1] == 'UNLOAD' or (row[1] == 'LOAD' and i == 1)))
                pressure.points.append(SignalPoint(ts, pressure_v))

        return [power, speed, pressure]


def get_power_param(segment: List[SignalPoint], flow: FlowCondition):
    sum_power = sum([pt.value for pt in segment])
    avg_power = sum_power / len(segment)
    return avg_power



import configparser
import os
from typing import List

from sha_learning.case_studies.energy_sim.sul_functions import label_event, parse_data, get_power_param, \
    is_chg_pt
from sha_learning.domain.lshafeatures import Event, NormalDistribution, Trace
from sha_learning.domain.sigfeatures import Timestamp, SampledSignal
from sha_learning.domain.sulfeatures import SystemUnderLearning, RealValuedVar, FlowCondition
from sha_learning.learning_setup.teacher import Teacher
from sha_learning.pltr.energy_pltr import double_plot

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

SPEED_RANGE = int(config['ENERGY CS']['SPEED_RANGE'])
MIN_SPEED = int(config['ENERGY CS']['MIN_SPEED'])
MAX_SPEED = int(config['ENERGY CS']['MAX_SPEED'])


def pwr_model(interval: List[Timestamp], P_0):
    interval = [ts.to_secs() for ts in interval]
    AVG_PW = 1.0
    return [AVG_PW] * len(interval)


# define flow conditions
on_fc: FlowCondition = FlowCondition(0, pwr_model)

# define distributions
off_distr = NormalDistribution(0, 0.0, 0.0)

model2distr = {0: []}
power = RealValuedVar([on_fc], [], model2distr, label='P')

# define events
events: List[Event] = []
for i in range(MIN_SPEED, MAX_SPEED, SPEED_RANGE):
    if i < MAX_SPEED - SPEED_RANGE:
        new_guard = '{}<=w<{}'.format(i, i + SPEED_RANGE)
    else:
        new_guard = '{}<=w'.format(i)
    events.append(Event(new_guard, 'start', 'm_{}'.format(len(events))))

spindle_off = Event('', 'stop', 'i_0')

events.append(spindle_off)

events.append(Event('', 'load', 'l'))
events.append(Event('', 'unload', 'u'))

DRIVER_SIG = ['w', 'pr']
DEFAULT_M = 0
DEFAULT_DISTR = 0

args = {'name': 'energy', 'driver': DRIVER_SIG, 'default_m': DEFAULT_M, 'default_d': DEFAULT_DISTR}
energy_sim_cs = SystemUnderLearning([power], events, parse_data, label_event, get_power_param, is_chg_pt, args=args)

test = False
if test:
    TEST_PATH = '/Users/lestingi/PycharmProjects/lsha/resources/traces/simulations/ENERGY/'
    traces_files = os.listdir(TEST_PATH)
    traces_files = [file for file in traces_files if file.startswith('_')]

    for file in traces_files:
        # testing data to signals conversion
        new_signals: List[SampledSignal] = parse_data(TEST_PATH + file)
        # testing chg pts identification
        chg_pts = energy_sim_cs.find_chg_pts([sig for sig in new_signals if sig.label in DRIVER_SIG])
        # testing event labeling
        id_events = [label_event(events, new_signals, pt.t) for pt in chg_pts[:10]]
        # testing signal to trace conversion
        energy_sim_cs.process_data(TEST_PATH + file)
        trace = energy_sim_cs.timed_traces[-1]
        print(Trace(tt=trace))
        power_pts = new_signals[0].points
        speed_pts = new_signals[1].points
        pressure_pts = new_signals[2].points
        double_plot([pt.timestamp for pt in power_pts], [pt.value for pt in power_pts],
                    [pt.timestamp for pt in speed_pts], [pt.value for pt in speed_pts],
                    trace, title=file, filtered=True,
                    timestamps3=[pt.timestamp for pt in pressure_pts],
                    v3=[pt.value for pt in pressure_pts])

    # test segment identification
    test_trace = Trace(energy_sim_cs.traces[0][:1])
    segments = energy_sim_cs.get_segments(test_trace)

    # test model identification
    TEACHER = Teacher(energy_sim_cs)
    identified_model: FlowCondition = TEACHER.mi_query(test_trace)
    print(identified_model)

    # test distr identification
    for i, trace in enumerate(TEACHER.timed_traces):
        for j, event in enumerate(trace.e):
            test_trace = Trace(energy_sim_cs.traces[i][:j])
            identified_distr = TEACHER.ht_query(test_trace, identified_model, save=True)

            segments = energy_sim_cs.get_segments(test_trace)
            avg_metrics = sum([TEACHER.sul.get_ht_params(segment, identified_model)
                               for segment in segments]) / len(segments)

            try:
                print('{}:\t{:.3f}->{}'.format(test_trace.events[-1].symbol, avg_metrics, identified_distr.params))
            except IndexError:
                print('{}:\t{:.3f}->{}'.format(test_trace, avg_metrics, identified_distr.params))



import configparser
import os
from typing import List, Dict

import numpy as np
import scipy.stats as stats
from tqdm import tqdm

from sha_learning.domain.lshafeatures import TimedTrace, FlowCondition, ProbDistribution, Trace
from sha_learning.domain.obstable import ObsTable, Row, State
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp
from sha_learning.domain.sulfeatures import SystemUnderLearning
from sha_learning.learning_setup.fastddtw import fast_ddtw, plot_aligned_signals
from sha_learning.learning_setup.logger import Logger
from sha_learning.learning_setup.trace_gen import TraceGenerator

LOGGER = Logger('TEACHER')

config = configparser.ConfigParser()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS = config['SUL CONFIGURATION']['CASE_STUDY']
NOISE = float(config['LSHA PARAMETERS']['DELTA'])
P_VALUE = 0.0
MI_QUERY = config['LSHA PARAMETERS']['MI_QUERY'] == 'True'
PLOT_DDTW = config['LSHA PARAMETERS']['PLOT_DDTW'] == 'True'
HT_QUERY = config['LSHA PARAMETERS']['HT_QUERY'] == 'True'
HT_QUERY_TYPE = config['LSHA PARAMETERS']['HT_QUERY_TYPE']


class Teacher:
    def __init__(self, sul: SystemUnderLearning, pov: str = None,
                 start_dt: str = None, end_dt: str = None, start_ts: int = None, end_ts: int = None):
        self.sul = sul

        # System-Dependent Attributes
        self.symbols = sul.symbols
        self.flows = sul.flows
        self.distributions = [v.distr for v in sul.vars]

        # Trace-Dependent Attributes
        self.timed_traces: List[TimedTrace] = sul.timed_traces
        self.signals: List[List[SampledSignal]] = sul.signals

        self.TG = TraceGenerator(pov=pov, start_dt=start_dt, end_dt=end_dt, start_ts=start_ts, end_ts=end_ts)
        self.hist = {}

    def add_distribution(self, d: ProbDistribution, f: FlowCondition):
        self.sul.add_distribution(d, f)

    # QUERIES
    @staticmethod
    def derivative(t: List[Timestamp], values: List[float]):
        # returns point-to-point increments for a given time-series
        # (derivative approximation)
        t = [x.to_secs() for x in t]
        increments = []
        try:
            increments = [(v - values[i - 1]) / (t[i] - t[i - 1]) for (i, v) in enumerate(values) if i > 0]
        except ZeroDivisionError:
            avg_dt = sum([x - t[i - 1] for (i, x) in enumerate(t) if i > 0]) / (len(t) - 1)
            increments = [(v - values[i - 1]) / avg_dt for (i, v) in enumerate(values) if i > 0]
        finally:
            return increments

    #############################################
    # MODEL IDENTIFICATION QUERY:
    # for a given prefix (word), gets all corresponding segments
    # and returns the flow condition that best fits such segments
    # If not enough data are available to draw a conclusion, returns None
    #############################################
    def mi_query(self, word: Trace):
        if not MI_QUERY or word == '':
            return self.flows[0][self.sul.default_m]
        else:
            segments = self.sul.get_segments(word)
            if len(segments) > 0:
                if len(self.flows[0]) == 1:
                    return self.flows[0][0]
                if CS == 'THERMO' and word[-1].symbol == 'h_0':
                    return self.flows[0][2]

                fits = []
                for segment in segments:
                    if len(segment) < 3:
                        continue
                    interval = [pt.timestamp for pt in segment]
                    # observed values and (approximate) derivative
                    real_behavior = [pt.value for pt in segment]
                    min_distance = 10000
                    best_fit = None

                    # for each model from the given input set
                    for flow in self.flows[0]:
                        ideal_model = flow.f(interval, segment[0].value)
                        # applies DDTW
                        res = fast_ddtw(real_behavior, ideal_model)

                        if PLOT_DDTW:
                            plot_aligned_signals(real_behavior, ideal_model, res[1])

                        if res[0] < min_distance:
                            min_distance = res[0]
                            best_fit = flow
                    else:
                        fits.append(best_fit)

                unique_fits = set(fits)
                freq = -1
                best_fit = None
                for f in unique_fits:
                    matches = sum([x == f for x in fits]) / len(fits)
                    if matches > freq:
                        freq = matches
                        best_fit = f
                if freq > 0.75:
                    return best_fit
                else:
                    LOGGER.info("!! INCONSISTENT PHYSICAL BEHAVIOR !!")
                    return None
            else:
                return None

    #############################################
    # HYPOTHESIS TESTING QUERY:
    # for a given prefix (word), gets all corresponding segments
    # and returns the random variable that best fits the randomly
    # generated model parameters.
    # If none of the available rand. variables fits the set of segments,
    # a new one is added
    # If available data are not enough to draw a conclusion, returns None
    #############################################
    def to_hist(self, values: List[float], d_id: int, update=False):
        try:
            if d_id in self.hist:
                distr = [d for d in self.distributions[0] if d.d_id == d_id][0]
                old_avg = distr.params['avg']
                old_v = len(self.hist[d_id])
                self.hist[d_id].extend(values)
            else:
                old_avg = 0.0
                old_v = 0
                self.hist[d_id] = values
            distr = [d for d in self.distributions[0] if d.d_id == d_id][0]
            distr.params['avg'] = (old_avg * old_v + sum(values)) / (old_v + len(values))
        except AttributeError:
            self.hist: Dict[int, List[float]] = {d.d_id: [] for d in self.distributions[0]}
            self.hist[d_id] = values

    def ht_query(self, word: Trace, flow: FlowCondition, save=True):
        if flow is None:
            return None

        if not HT_QUERY or word == '':
            return self.distributions[self.sul.default_d]

        if HT_QUERY_TYPE == 'D':
            return self.ht_d_query(word, flow, save)
        else:
            return self.ht_s_query(word, flow, save)

    def ht_d_query(self, word: Trace, flow: FlowCondition, save=True):
        segments = self.sul.get_segments(word)
        if len(segments) > 0:
            eligible_distributions = self.sul.vars[0].get_distr_for_flow(flow.f_id)

            metrics = [self.sul.get_ht_params(segment, flow) for segment in segments]
            metrics = [met for met in metrics if met is not None]
            unique_metrics = list(set(metrics))
            if len(unique_metrics) > 1:
                LOGGER.error('INCONSISTENT PHYSICAL BEHAVIOR')
                raise RuntimeError

            best_fit: ProbDistribution = None

            try:
                for distr in self.hist:
                    value = self.hist[distr][0]
                    fits = [e_d for e_d in eligible_distributions if e_d.d_id == distr]
                    if value == unique_metrics[0] and len(fits) > 0:
                        best_fit = fits[0]
                        break
            except AttributeError:
                pass

            if best_fit is None:
                new_distr = ProbDistribution(len(self.distributions[0]), {'avg': unique_metrics[0]})
                if save:
                    self.add_distribution(new_distr, flow)
                    self.to_hist(metrics, new_distr.d_id)
                return new_distr
            else:
                self.to_hist(metrics, best_fit.d_id, update=True)
                return best_fit

    def ht_s_query(self, word: Trace, flow: FlowCondition, save=True):
        segments = self.sul.get_segments(word)
        if len(segments) > 0:
            # distr associated with selected flow
            eligible_distributions = self.sul.vars[0].get_distr_for_flow(flow.f_id)

            # randomly distributed metrics for each segment
            metrics = [self.sul.get_ht_params(segment, flow) for segment in segments]
            metrics = [met for met in metrics if met is not None]
            avg_metrics = sum(metrics) / len(metrics)

            min_dist, best_fit = 1000, None

            try:
                for distr in self.hist:
                    if len(self.hist[distr]) == 0 or len(metrics) == 0:
                        continue

                    if CS == 'THERMO':
                        v1 = metrics
                        noise1 = [0] * len(v1)
                    else:
                        v1 = [avg_metrics] * 50
                        noise1 = np.random.normal(0.0, NOISE, size=len(v1))

                    v1 = [x + noise1[i] for i, x in enumerate(v1)]

                    v2 = []
                    if CS == 'THERMO':
                        v2 = self.hist[distr]
                        noise2 = [0] * len(v2)
                    else:
                        for m in self.hist[distr]:
                            v2 += [m] * 10
                        noise2 = np.random.normal(0.0, NOISE, size=len(v2))
                    v2 = [x + noise2[i] for i, x in enumerate(v2)]

                    statistic, pvalue = stats.ks_2samp(v1, v2)
                    fits = [d for d in eligible_distributions if d.d_id == distr]
                    if statistic <= min_dist and pvalue >= P_VALUE and len(fits) > 0:
                        min_dist = statistic
                        best_fit = fits[0]
            except AttributeError:
                pass

            if best_fit is not None and min_dist < 1.0:
                self.to_hist(metrics, best_fit.d_id, update=True)
                return best_fit
            else:
                new_distr = ProbDistribution(len(self.distributions[0]), {'avg': sum(metrics) / len(metrics)})
                if save:
                    self.add_distribution(new_distr, flow)
                    self.to_hist(metrics, new_distr.d_id)
                return new_distr

    #############################################
    # ROW EQUALITY QUERY:
    # checks if two rows (row(s1), row(s2)) are weakly equal
    # returns true/false
    #############################################
    def eqr_query(self, row1: Row, row2: Row, strict=False):
        if strict:
            return row1 == row2

        for i, state in enumerate(row1.state):
            # if both rows have filled cells which differ from each other,
            # weak equality is violated
            if state.observed() and row2.state[i].observed() and state != row2.state[i]:
                return False
        else:
            return True

    #############################################
    # KNOWLEDGE REFINEMENT QUERY:
    # checks if there are ambiguous words in the observation table
    # if so, it samples new traces (through the TraceGenerator)
    # to gain more knowledge about the system under learning
    #############################################
    def ref_query(self, table: ObsTable):
        LOGGER.info('Performing ref query...')

        n_resample = int(config['LSHA PARAMETERS']['N_min'])
        S = table.get_S()
        upp_obs: List[Row] = table.get_upper_observations()
        lS = table.get_low_S()
        low_obs: List[Row] = table.get_lower_observations()

        # find all words which are ambiguous
        # (equivalent to multiple rows)
        amb_words: List[Trace] = []
        for i, row in tqdm(enumerate(upp_obs + low_obs)):
            # if there are not enough observations of a word,
            # it needs a refinement query
            s = S[i] if i < len(upp_obs) else lS[i - len(upp_obs)]
            for e_i, e in enumerate(table.get_E()):
                if len(self.sul.get_segments(s + e)) < n_resample:
                    amb_words.append(s + e)

            if not row.is_populated():
                continue

            # find equivalent rows
            eq_rows: List[Row] = []
            for (j, row_2) in enumerate(upp_obs):
                if row_2.is_populated() and i != j and self.eqr_query(row, row_2):
                    eq_rows.append(row_2)
            if len(set(eq_rows)) > 1:
                amb_words.append(s)

        # sample new traces only for ambiguous words which
        # are not prefixes of another ambiguous word
        uq = amb_words
        # for i, w in tqdm(enumerate(amb_words)):
        #     suffixes = [w2 for w2 in amb_words if w2 != w and w2.startswith(w)]
        #     if len(suffixes) == 0:
        #         uq.append(w)

        for word in tqdm(uq, total=len(uq)):
            LOGGER.info('Requesting new traces for {}'.format(str(word)))
            for e in table.get_E():
                self.TG.set_word(word + e)
                path = self.TG.get_traces(n_resample)
                if path is not None:
                    for sim in path:
                        self.sul.process_data(sim)
                else:
                    LOGGER.debug('!! An error occurred while generating traces !!')

    #############################################
    # COUNTEREXAMPLE QUERY:
    # looks for a counterexample to current obs. table
    # returns counterexample t if:
    # -> t highlights non-closedness
    # -> t highlights non-consistency
    #############################################
    def get_counterexample(self, table: ObsTable):
        LOGGER.info('Looking for counterexample...')

        if CS == 'THERMO' and len(self.timed_traces) >= 2000:
            return None

        S = table.get_S()
        low_S = table.get_low_S()

        traces: List[Trace] = self.sul.traces
        not_counter: List[Trace] = []
        for i, trace in tqdm(enumerate(traces), total=len(traces)):
            for prefix in trace.get_prefixes():
                LOGGER.debug('Checking {}'.format(str(prefix)))
                # prefix is still not in table
                if prefix not in S and prefix not in low_S and prefix not in not_counter:
                    # fills hypothetical new row
                    new_row = Row([])
                    for e_i, e_word in enumerate(table.get_E()):
                        word = prefix + e_word
                        id_model = self.mi_query(word)
                        id_distr = self.ht_query(word, id_model, save=False)
                        if id_model is not None and id_distr is not None:
                            new_row.state.append(State([(id_model, id_distr)]))
                        else:
                            new_row.state.append(State([(None, None)]))
                    # if there are sufficient data to fill the new row
                    if new_row.is_populated():
                        eq_rows = [row for row in table.get_upper_observations() if self.eqr_query(new_row, row)]
                        not_ambiguous = len(set(eq_rows)) <= 1
                        diff_rows = [row for row in eq_rows if
                                     not self.eqr_query(new_row, row, strict=True)]

                        if len(diff_rows) > 0:
                            # found non-closedness
                            LOGGER.warn("!! MISSED NON-CLOSEDNESS !!")
                            return prefix
                        elif not_ambiguous:
                            # checks non-consistency only for rows that are not ambiguous
                            for s_i, s_word in enumerate(S):
                                old_row = table.get_upper_observations()[s_i] if s_i < len(S) else \
                                    table.get_lower_observations()[s_i - len(S)]
                                # finds weakly equal rows in S
                                if self.eqr_query(old_row, new_row):
                                    for event in self.sul.events:
                                        # if the hypothetical discriminating event is already in E
                                        discr_is_prefix = False
                                        for e in table.get_E():
                                            if str(e).startswith(event.symbol):
                                                continue
                                        # else checks all 1-step distant rows
                                        if s_word + Trace([event]) in S:
                                            old_row_a: Row = table.get_upper_observations()[
                                                S.index(s_word + Trace([event]))]
                                        elif s_word + Trace([event]) in low_S:
                                            old_row_a: Row = table.get_lower_observations()[
                                                low_S.index(s_word + Trace([event]))]
                                        else:
                                            continue
                                        row_1_filled = old_row_a.state[0].observed()
                                        row_2 = Row([])
                                        for e in table.get_E():
                                            id_model_2 = self.mi_query(prefix + Trace([event]) + e)
                                            id_distr_2 = self.ht_query(prefix + Trace([event]) + e, id_model_2,
                                                                       save=False)
                                            if id_model_2 is None or id_distr_2 is None:
                                                row_2.state.append(State([(None, None)]))
                                            else:
                                                row_2.state.append(State([(id_model_2, id_distr_2)]))
                                        row_2_filled = row_2.state[0].observed()
                                        if row_1_filled and row_2_filled and not discr_is_prefix and \
                                                not self.eqr_query(row_2, old_row_a):
                                            LOGGER.warn(
                                                "!! MISSED NON-CONSISTENCY ({}, {}) !!".format(Trace([event]), s_word))
                                            return prefix
                                    else:
                                        not_counter.append(prefix)
                        else:
                            not_counter.append(prefix)
        else:
            if CS in ['ENERGY', 'AUTO_TWIN'] and len(not_counter) > 0:
                new_events = set([e.symbol for x in not_counter for e in x.events]) - \
                             set([e.symbol for t in S for e in t.events])
                if len(new_events) > 0:  # or not_counter[-1] not in S:
                    return not_counter[-1]
                else:
                    return None
            else:
                return None



import configparser
import os
from enum import Enum

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()


class LogLevel(Enum):
    INFO = 1
    DEBUG = 2
    WARNING = 3
    ERROR = 4
    MSG = 99

    def __str__(self):
        if self.value == 1:
            return 'INFO'
        elif self.value == 2:
            return 'DEBUG'
        elif self.value == 3:
            return 'WARNING'
        elif self.value == 4:
            return 'ERROR'
        elif self.value == 99:
            return 'MSG'
        else:
            return ''

    @staticmethod
    def parse_str(s):
        if s == 'INFO':
            return LogLevel.INFO
        elif s == 'DEBUG':
            return LogLevel.DEBUG
        elif s == 'WARNING':
            return LogLevel.WARNING
        elif s == 'ERROR':
            return LogLevel.ERROR
        elif s == 'MSG':
            return LogLevel.MSG
        else:
            return None


# INIT LOGGING LEVEL BASED ON CONFIG FILE
if 'LoggingLevel' in config['DEFAULT']:
    MIN_LOG_LEVEL = LogLevel.parse_str(config['DEFAULT']['LoggingLevel']).value
else:
    MIN_LOG_LEVEL = LogLevel.WARNING.value


#

class Logger:
    def __init__(self, speaker: str):
        self.speaker = speaker
        self.format = "\n[{}] ({})\t{}"
        pass

    def info(self, msg):
        if MIN_LOG_LEVEL <= LogLevel.INFO.value:
            print(self.format.format(self.speaker, str(LogLevel.INFO), msg))

    def debug(self, msg):
        if MIN_LOG_LEVEL <= LogLevel.DEBUG.value:
            print(self.format.format(self.speaker, str(LogLevel.DEBUG), msg))

    def warn(self, msg):
        if MIN_LOG_LEVEL <= LogLevel.WARNING.value:
            print(self.format.format(self.speaker, str(LogLevel.WARNING), msg))

    def error(self, msg):
        if MIN_LOG_LEVEL <= LogLevel.ERROR.value:
            print(self.format.format(self.speaker, str(LogLevel.ERROR), msg))

    def msg(self, msg):
        if MIN_LOG_LEVEL <= LogLevel.MSG.value:
            print(self.format.format(self.speaker, str(LogLevel.MSG), msg))



import configparser
import math
import os
from typing import List

from sha_learning.domain.lshafeatures import FlowCondition
from sha_learning.domain.sigfeatures import SampledSignal, Timestamp, Event, SignalPoint
from sha_learning.learning_setup.logger import Logger

config = configparser.ConfigParser()
config.sections()
config.read(
    os.path.dirname(os.path.abspath(__file__)).split('sha_learning')[0] + 'sha_learning/resources/config/config.ini')
config.sections()

CS_VERSION = int(config['SUL CONFIGURATION']['CS_VERSION'].replace('\n', '')[0])
SAMPLE_STRATEGY = config['SUL CONFIGURATION']['RESAMPLE_STRATEGY']
LOGGER = Logger('SUL DATA HANDLER')


def is_chg_pt(curr, prev):
    return curr != prev


def label_event(events: List[Event], signals: List[SampledSignal], t: Timestamp):
    posX = signals[1]
    moving = signals[2]

    '''
    Repeat for every channel in the system
    '''
    curr_mov = list(filter(lambda x: x.timestamp == t, moving.points))[0]
    identified_channel = 'start' if curr_mov.value == 1 else 'stop'

    identified_guard = ''
    if (SAMPLE_STRATEGY == 'SIM' and CS_VERSION in [1, 2, 3]) or \
            (SAMPLE_STRATEGY == 'UPPAAL' and CS_VERSION in [2, 3, 4, 5]):
        posY = signals[3]
        curr_posx = list(filter(lambda x: x.timestamp <= t, posX.points))[-1]
        curr_posy = list(filter(lambda x: x.timestamp <= t, posY.points))[-1]
        if SAMPLE_STRATEGY == 'UPPAAL':
            in_waiting = curr_posx.value >= 2000.0 and curr_posy.value <= 3000.0
        else:
            in_waiting = 16 <= curr_posx.value <= 23.0 and 1.0 <= curr_posy.value <= 10.0
        identified_guard += 'sit' if in_waiting else '!sit'

    if (SAMPLE_STRATEGY == 'SIM' and CS_VERSION in [1, 2, 3]) or \
            (SAMPLE_STRATEGY == 'UPPAAL' and CS_VERSION in [3, 4, 5]):
        posY = signals[3]
        curr_posx = list(filter(lambda x: x.timestamp <= t, posX.points))[-1]
        curr_posy = list(filter(lambda x: x.timestamp <= t, posY.points))[-1]
        if SAMPLE_STRATEGY == 'UPPAAL':
            in_office = curr_posx.value >= 2000.0 and 1000.0 <= curr_posy.value <= 3000.0
        else:
            in_office = 1.0 <= curr_posx.value <= 11.0 and 1.0 <= curr_posy.value <= 10.0
        identified_guard += 'run' if in_office else '!run'

    if SAMPLE_STRATEGY == 'SIM' and CS_VERSION in [4]:
        posY = signals[3]
        curr_posx = list(filter(lambda x: x.timestamp <= t, posX.points))[-1]
        curr_posy = list(filter(lambda x: x.timestamp <= t, posY.points))[-1]

        close_to_chair = 16 <= curr_posx.value <= 20.0 and 3.0 <= curr_posy.value <= 6.0
        identified_guard += 's' if close_to_chair and identified_channel == 'stop' else '!s'

        next_sig_x = list(filter(lambda x: x.timestamp > t, posX.points))
        next_sig_y = list(filter(lambda x: x.timestamp > t, posY.points))
        nextx = next_sig_x[0] if len(next_sig_x) > 0 else curr_posx
        nexty = next_sig_y[0] if len(next_sig_y) > 0 else curr_posy
        dist = math.sqrt((nextx.value - curr_posx.value) ** 2 + (nexty.value - curr_posy.value) ** 2)
        vel = dist / 2.0

        room = signals[4]
        curr_room_status = list(filter(lambda x: x.timestamp <= t.to_secs(), room.points))[-1]
        identified_guard += 'r' if vel > 0.8 and ((close_to_chair and identified_channel == 'stop')
                                                  or curr_room_status.value) else '!r'
        identified_guard += 'h' if curr_room_status.value and not (close_to_chair and vel > 1.0) else '!h'

        identified_guard += 'l' if 0.2 <= vel < 0.4 and not close_to_chair \
                                   and (identified_channel == 'start' or curr_room_status.value) else '!l'
        identified_guard += 'a' if vel < 0.2 and not curr_room_status else '!a'

    '''
    Find symbol associated with guard-channel combination
    '''
    identified_event = [e for e in events if e.guard == identified_guard and e.chan == identified_channel][0]
    return identified_event


def parse_data(path: str):
    if SAMPLE_STRATEGY == 'SIM':
        return parse_traces_sim(path)
    else:
        return parse_traces_uppaal(path)


def parse_traces_sim(path: str):
    if CS_VERSION == 4:
        logs = ['humanFatigue.log', 'humanPosition.log', 'environmentData.log']
    else:
        logs = ['humanFatigue.log', 'humanPosition.log']

    new_traces: List[SampledSignal] = []
    for i, log in enumerate(logs):
        f = open(path + log)
        lines = f.readlines()[1:]
        lines = [line.replace('\n', '') for line in lines]
        t = [float(line.split(':')[0]) for line in lines]
        if i == 0:
            v = [float(line.split(':')[2]) for line in lines]
            signal = [SignalPoint(Timestamp(0, 0, 0, 0, 0, x), v[j]) for j, x in enumerate(t)]
            new_traces.append(SampledSignal(signal, str(len(new_traces))))
        elif i == 1:
            pos = [line.split(':')[2] for line in lines]
            pos_x = [float(line.split('#')[0]) for line in pos]
            signal = [SignalPoint(Timestamp(0, 0, 0, 0, 0, x), pos_x[j]) for j, x in enumerate(t)]
            new_traces.append(SampledSignal(signal, 'humanPositionX'))
            busy = [float(v != pos[j - 1]) for j, v in enumerate(pos) if j > 0]
            busy = [0.0] + busy
            signal = [SignalPoint(Timestamp(0, 0, 0, 0, 0, x), busy[j]) for j, x in enumerate(t)]
            new_traces.append(SampledSignal(signal, str(len(new_traces))))
            pos_y = [float(line.split('#')[1]) for line in pos]
            signal = [SignalPoint(Timestamp(0, 0, 0, 0, 0, x), pos_y[j]) for j, x in enumerate(t)]
            new_traces.append(SampledSignal(signal, 'humanPositionY'))
        else:
            data = [line.split(':')[1] for line in lines]
            data = list(map(lambda x: (float(x.split('#')[0]), float(x.split('#')[1])), data))
            harsh = []
            for pt in data:
                temp = pt[0]
                hum = pt[1]
                harsh.append(temp <= 12.0 or temp >= 32.0 or hum <= 30.0 or hum >= 60.0)
            signal = [SignalPoint(x, harsh[j]) for (j, x) in enumerate(t)]
            new_traces.append(SampledSignal(signal, log))

    return new_traces


def parse_traces_uppaal(path: str):
    f = open(path, 'r')
    if CS_VERSION in [1, 2]:
        variables = ['humanFatigue[currH - 1]', 'humanPositionX[currH - 1]',
                     'amy.busy || amy.p_2', 'humanPositionY[currH - 1]']
    else:
        variables = ['humanFatigue[currH - 1]', 'humanPositionX[currH - 1]',
                     'amy.busy || amy.p_2 || amy.run || amy.p_4', 'humanPositionY[currH - 1]']
    lines = f.readlines()
    split_indexes = [lines.index(k + ':\n') for k in variables]
    split_lines = [lines[i + 1:split_indexes[ind + 1]] for (ind, i) in enumerate(split_indexes) if
                   i != split_indexes[-1]]
    split_lines.append(lines[split_indexes[-1] + 1:len(lines)])
    traces = len(split_lines[0])
    new_traces: List[SampledSignal] = []
    for trace in range(traces):
        for i, v in enumerate(variables):
            entries = split_lines[i][trace].split(' ')
            entries = entries[1:]
            for e in entries:
                new = e.replace('(', '')
                new = new.replace(')', '')
                entries[entries.index(e)] = new
            t = [float(x.split(',')[0]) for x in entries]
            v = [float(x.split(',')[1]) for x in entries]
            signal = [SignalPoint(Timestamp(0, 0, 0, 0, 0, t[i]), v[i]) for i in range(len(t))]
            new_traces.append(SampledSignal(signal, str(i)))
    return new_traces


def get_ftg_param(segment: List[SignalPoint], flow: FlowCondition):
    try:
        val = [pt.value for pt in segment]
        # metric for walking
        if flow.f_id == 1:
            lambdas = []
            for (i, v) in enumerate(val):
                if i > 0 and v != val[i - 1]:
                    lambdas.append(math.log((1 - v) / (1 - val[i - 1])))
            est_rate = sum(lambdas) / len(lambdas) if len(lambdas) > 0 else None
        # metric for standing/sitting
        else:
            mus = []
            for (i, v) in enumerate(val):
                if i > 0 and v != val[i - 1] and val[i - 1] != 0:
                    mus.append(math.log(v / val[i - 1]))
            est_rate = sum(mus) / len(mus) if len(mus) > 0 else None

        return abs(est_rate) if est_rate is not None else None
    except ValueError:
        return None
